@@ -2,7 +2,7 @@
  * ProFTPD: mod_sql_postgres -- Support for connecting to Postgres databases.
  * Time-stamp: <1999-10-04 03:21:21 root>
  * Copyright (c) 2001 Andrew Houghton
- * Copyright (c) 2004 TJ Saunders
+ * Copyright (c) 2004-2007 TJ Saunders
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  * the resulting executable, without including the source code for OpenSSL in
  * the source distribution.
  *
- * $Id: mod_sql_postgres.c,v 1.27 2005/04/11 17:02:33 castaglia Exp $
+ * $Id: mod_sql_postgres.c,v 1.32 2007/07/06 22:40:58 castaglia Exp $
  */
 
 /*
@@ -34,18 +34,19 @@
  * Internal define used for debug and logging.  All backends are encouraged
  * to use the same format.
  */
-#define MOD_SQL_POSTGRES_VERSION	"mod_sql_postgres/4.02"
+#define MOD_SQL_POSTGRES_VERSION	"mod_sql_postgres/4.03"
 
 #define _POSTGRES_PORT "5432"
 
-#include <libpq-fe.h>
 #include "conf.h"
 #include "../contrib/mod_sql.h"
 
+#include <libpq-fe.h>
+
 /* 
  * timer-handling code adds the need for a couple of forward declarations
  */
-MODRET cmd_close( cmd_rec *cmd );
+MODRET cmd_close(cmd_rec *cmd);
 module sql_postgres_module;
 
 /* 
@@ -198,9 +199,9 @@ static int _sql_timer_callback(CALLBACK_FRAME) {
  */
 static modret_t *_build_error(cmd_rec *cmd, db_conn_t *conn) {
   if (!conn)
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
 
-  return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+  return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
     PQerrorMessage(conn->postgres));
 }
 
@@ -219,7 +220,7 @@ static modret_t *_build_data(cmd_rec *cmd, db_conn_t *conn) {
   int row =0;
 
   if (!conn) 
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
 
   result = conn->result;
 
@@ -269,14 +270,15 @@ MODRET cmd_open(cmd_rec *cmd) {
 
   if (cmd->argc < 1) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }    
 
   /* get the named connection */
 
   if (!(entry = _sql_get_connection(cmd->argv[0]))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   } 
 
   conn = (db_conn_t *) entry->data;
@@ -293,7 +295,7 @@ MODRET cmd_open(cmd_rec *cmd) {
     sql_log(DEBUG_INFO, "connection '%s' count is now %d", entry->name,
       entry->connections);
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-    return HANDLED(cmd);
+    return PR_HANDLED(cmd);
   }
 
   /* make sure we have a new conn struct */
@@ -328,7 +330,7 @@ MODRET cmd_open(cmd_rec *cmd) {
     entry->connections);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_open");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -365,13 +367,14 @@ MODRET cmd_close(cmd_rec *cmd) {
 
   if ((cmd->argc < 1) || (cmd->argc > 2)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   if (!(entry = _sql_get_connection(cmd->argv[0]))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -382,7 +385,7 @@ MODRET cmd_close(cmd_rec *cmd) {
       entry->connections);
 
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
-    return HANDLED(cmd);
+    return PR_HANDLED(cmd);
   }
 
   /* decrement connections. If our count is 0 or we received a second arg
@@ -407,7 +410,7 @@ MODRET cmd_close(cmd_rec *cmd) {
     entry->connections);
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_close");
   
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -454,7 +457,21 @@ MODRET cmd_defineconnection(cmd_rec *cmd) {
 
   if ((cmd->argc < 4) || (cmd->argc > 5) || (!cmd->argv[0])) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+  }
+
+  if (!conn_pool) {
+    pr_log_pri(PR_LOG_WARNING, "warning: the mod_sql_postgres module has not "
+      "been properly initialized.  Please make sure your --with-modules "
+      "configure option lists mod_sql *before* mod_sql_postgres, and "
+      "recompile.");
+      
+    sql_log(DEBUG_FUNC, "%s", "The mod_sql_postgres module has not been "
+      "properly initialized.  Please make sure your --with-modules configure "
+      "option lists mod_sql *before* mod_sql_postgres, and recompile.");
+    sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
+
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "uninitialized module");
   }
 
   conn = (db_conn_t *) palloc(conn_pool, sizeof(db_conn_t));
@@ -506,7 +523,7 @@ MODRET cmd_defineconnection(cmd_rec *cmd) {
   /* insert the new conn_info into the connection hash */
   if (!(entry = _sql_add_connection(conn_pool, name, (void *) conn))) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
       "named connection already exists");
   }
 
@@ -526,7 +543,7 @@ MODRET cmd_defineconnection(cmd_rec *cmd) {
   sql_log(DEBUG_INFO, "  ttl: '%d'", entry->ttl);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_defineconnection");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -555,7 +572,7 @@ static modret_t *cmd_exit(cmd_rec *cmd) {
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_exit");
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -619,14 +636,15 @@ MODRET cmd_select(cmd_rec *cmd) {
 
   if (cmd->argc < 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_select");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_select");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
   
   conn = (db_conn_t *) entry->data;
@@ -756,14 +774,15 @@ MODRET cmd_insert(cmd_rec *cmd) {
 
   if ((cmd->argc != 2) && (cmd->argc != 4)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_insert");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_insert");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -811,7 +830,7 @@ MODRET cmd_insert(cmd_rec *cmd) {
   SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_insert");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -858,14 +877,15 @@ MODRET cmd_update(cmd_rec *cmd) {
 
   if ((cmd->argc < 2) || (cmd->argc > 4)) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_update");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_update");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -914,7 +934,7 @@ MODRET cmd_update(cmd_rec *cmd) {
   SQL_FREE_CMD(close_cmd);
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_update");
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /*
@@ -942,14 +962,14 @@ MODRET cmd_procedure(cmd_rec *cmd) {
 
   if (cmd->argc != 3) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_procedure");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* PostgreSQL supports procedures, but the backend doesn't. */
 
   sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_procedure");
 
-  return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+  return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
     "backend does not support procedures");
 }
 
@@ -986,14 +1006,15 @@ MODRET cmd_query(cmd_rec *cmd) {
 
   if (cmd->argc != 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_query");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_query");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1040,7 +1061,7 @@ MODRET cmd_query(cmd_rec *cmd) {
       sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_query");
     }
   } else {
-    dmr = HANDLED(cmd);
+    dmr = PR_HANDLED(cmd);
   }
 
   /* close the connection, return the data. */
@@ -1090,14 +1111,15 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
 
   if (cmd->argc != 2) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_escapestring");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1126,12 +1148,12 @@ MODRET cmd_escapestring(cmd_rec * cmd) {
  *  cmd->argv[2]: hashed string
  *
  * Returns:
- *  HANDLED(cmd)                        -- passwords match
- *  ERROR_INT(cmd, PR_AUTH_NOPWD)       -- missing password
- *  ERROR_INT(cmd, PR_AUTH_BADPWD)      -- passwords don't match
- *  ERROR_INT(cmd, PR_AUTH_DISABLEPWD)  -- password is disabled
- *  ERROR_INT(cmd, PR_AUTH_AGEPWD)      -- password is aged
- *  ERROR(cmd)                          -- unknown error
+ *  PR_HANDLED(cmd)                        -- passwords match
+ *  PR_ERROR_INT(cmd, PR_AUTH_NOPWD)       -- missing password
+ *  PR_ERROR_INT(cmd, PR_AUTH_BADPWD)      -- passwords don't match
+ *  PR_ERROR_INT(cmd, PR_AUTH_DISABLEPWD)  -- password is disabled
+ *  PR_ERROR_INT(cmd, PR_AUTH_AGEPWD)      -- password is aged
+ *  PR_ERROR(cmd)                          -- unknown error
  *
  * Notes:
  *  If this backend does not provide this functionality, this cmd *must*
@@ -1147,14 +1169,15 @@ MODRET cmd_checkauth(cmd_rec * cmd) {
 
   if (cmd->argc != 3) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_checkauth");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "badly formed request");
   }
 
   /* get the named connection -- not used in this case, but for consistency */
   entry = _sql_get_connection(cmd->argv[0]);
   if (!entry) {
     sql_log(DEBUG_FUNC, "%s", "exiting \tpostgres cmd_checkauth");
-    return ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION, "unknown named connection");
+    return PR_ERROR_MSG(cmd, MOD_SQL_POSTGRES_VERSION,
+      "unknown named connection");
   }
 
   conn = (db_conn_t *) entry->data;
@@ -1163,7 +1186,7 @@ MODRET cmd_checkauth(cmd_rec * cmd) {
 
   /* PostgreSQL doesn't provide this functionality */
 
-  return ERROR(cmd);
+  return PR_ERROR(cmd);
 }
 
 /*
@@ -1203,24 +1226,46 @@ MODRET cmd_identify(cmd_rec * cmd) {
   return mod_create_data(cmd, (void *) sd);
 }  
 
+MODRET cmd_prepare(cmd_rec *cmd) {
+  if (cmd->argc != 1) {
+    return PR_ERROR(cmd);
+  }
+
+  conn_pool = (pool *) cmd->argv[0];
+  conn_cache = make_array((pool *) cmd->argv[0], DEF_CONN_POOL_SIZE,
+    sizeof(conn_entry_t));
+
+  return mod_create_data(cmd, NULL);
+}
+
+MODRET cmd_cleanup(cmd_rec *cmd) {
+  destroy_pool(conn_pool);
+  conn_pool = NULL;
+  conn_cache = NULL;
+
+  return mod_create_data(cmd, NULL);
+}
+
 /* SQL cmdtable: mod_sql requires each backend module to define a cmdtable
  *  with this exact name. ALL these functions must be defined; mod_sql checks
  *  that they all exist on startup and ProFTPD will refuse to start if they
  *  aren't defined.
  */
 static cmdtable sql_postgres_cmdtable[] = {
-  { CMD, "sql_open",             G_NONE, cmd_open,             FALSE, FALSE },
+  { CMD, "sql_checkauth",        G_NONE, cmd_checkauth,        FALSE, FALSE },
+  { CMD, "sql_cleanup",          G_NONE, cmd_cleanup,          FALSE, FALSE },
   { CMD, "sql_close",            G_NONE, cmd_close,            FALSE, FALSE },
   { CMD, "sql_defineconnection", G_NONE, cmd_defineconnection, FALSE, FALSE },
+  { CMD, "sql_escapestring",     G_NONE, cmd_escapestring,     FALSE, FALSE },
   { CMD, "sql_exit",             G_NONE, cmd_exit,             FALSE, FALSE },
-  { CMD, "sql_select",           G_NONE, cmd_select,           FALSE, FALSE },
+  { CMD, "sql_identify",         G_NONE, cmd_identify,         FALSE, FALSE },
   { CMD, "sql_insert",           G_NONE, cmd_insert,           FALSE, FALSE },
-  { CMD, "sql_update",           G_NONE, cmd_update,           FALSE, FALSE },
+  { CMD, "sql_open",             G_NONE, cmd_open,             FALSE, FALSE },
+  { CMD, "sql_prepare",          G_NONE, cmd_prepare,          FALSE, FALSE },
   { CMD, "sql_procedure",        G_NONE, cmd_procedure,        FALSE, FALSE },
   { CMD, "sql_query",            G_NONE, cmd_query,            FALSE, FALSE },
-  { CMD, "sql_escapestring",     G_NONE, cmd_escapestring,     FALSE, FALSE },
-  { CMD, "sql_checkauth",        G_NONE, cmd_checkauth,        FALSE, FALSE },
-  { CMD, "sql_identify",         G_NONE, cmd_identify,         FALSE, FALSE },
+  { CMD, "sql_select",           G_NONE, cmd_select,           FALSE, FALSE },
+  { CMD, "sql_update",           G_NONE, cmd_update,           FALSE, FALSE },
 
   { 0, NULL }
 };
