@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2006 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,7 +27,7 @@
 /* Memory allocation/anti-leak system.  Yes, this *IS* stolen from Apache
  * also.  What can I say?  It makes sense, and it's safe (more overhead
  * though)
- * $Id: pool.h,v 1.18 2004/01/09 04:23:15 castaglia Exp $
+ * $Id: pool.h,v 1.20 2007/01/11 04:05:07 castaglia Exp $
  */
 
 #ifndef PR_POOL_H
@@ -42,13 +42,6 @@ void free_pools(void);
 pool *make_sub_pool(pool *);		/* All pools are sub-pools of perm */
 pool *pr_pool_create_sz(pool *, int);
 
-/* The make_named_sub_pool() function, which was never really supported,
- * has been deprecated.  This macro is for supporting those contrib modules
- * that still may make use of this old function.  Newer code should simply
- * use make_sub_pool() directly.
- */
-#define make_named_sub_pool(p, s)       make_sub_pool((p))
-
 /* Clears out _everything_ in a pool, destroying any sub-pools */
 void destroy_pool(struct pool *);
 
@@ -61,11 +54,12 @@ extern char *pstrdup(struct pool *, const char *);
 extern char *pstrndup(struct pool *, const char *, int);
 char *pstrcat(struct pool *, ...);       /* Must be char * */
 char *pdircat(struct pool *, ...);	/* Must be char * */
-
-/* MM debugging */
-void pr_pool_debug_memory(void (*)(const char *, ...));
 void pr_pool_tag(struct pool *, const char *);
 
+#ifdef PR_USE_DEVEL
+void pr_pool_debug_memory(void (*)(const char *, ...));
+#endif /* PR_USE_DEVEL */
+
 /* Array management */
 
 typedef struct {
@@ -90,16 +84,7 @@ array_header *copy_array_hdr(pool *, const array_header *);
 extern void pr_alarms_block(void);
 extern void pr_alarms_unblock(void);
 
-FILE *pfopen(struct pool *, const char *, const char *);
-FILE *pfdopen(struct pool *, int, const char *);
-int popenf(struct pool *, const char *, int, int);
-
-int pfclose(struct pool *, FILE *);
-int pclosef(struct pool *, int);
-
-/* Functions for cleanup handlers */
 void register_cleanup(pool *, void *, void (*)(void *), void (*)(void *));
-void register_file_cleanups(pool *, FILE *);
 void unregister_cleanup(pool *, void *, void (*)(void *));
 
 /* minimum free bytes in a new block pool */
