@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2007 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Non-specific support functions.
- * $Id: support.h,v 1.25 2005/02/26 17:28:58 castaglia Exp $
+ * $Id: support.h,v 1.28 2007/01/12 01:30:55 castaglia Exp $
  */
 
 #ifndef PR_SUPPORT_H
@@ -48,7 +48,6 @@
 int getopt(int, char * const [], const char *);
 extern char *optarg;
 extern int optind,opterr,optopt;
-#endif
 
 #ifndef HAVE_GETOPT_LONG
 struct option {
@@ -60,7 +59,8 @@ struct option {
 
 int getopt_long(int, char * const [], const char *, const struct option *,
   int *);
-#endif
+# endif /* !HAVE_GETOPT_LONG */
+#endif /* !HAVE_GETOPT */
 
 void pr_signals_block(void);
 void pr_signals_unblock(void);
