@@ -2,7 +2,7 @@
  * ProFTPD: mod_delay -- a module for adding arbitrary delays to the FTP
  *                       session lifecycle
  *
- * Copyright (c) 2004-2006 TJ Saunders
+ * Copyright (c) 2004-2007 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,13 +26,13 @@
  * This is mod_delay, contrib software for proftpd 1.2.10 and above.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_delay.c,v 1.19 2006/03/17 18:52:37 castaglia Exp $
+ * $Id: mod_delay.c,v 1.26 2007/10/01 21:41:28 castaglia Exp $
  */
 
 #include "conf.h"
 #include "privs.h"
 
-#define MOD_DELAY_VERSION		"mod_delay/0.5"
+#define MOD_DELAY_VERSION		"mod_delay/0.6"
 
 /* Make sure the version of proftpd is as necessary. */
 #if PROFTPD_VERSION_NUMBER < 0x0001021001
@@ -45,6 +45,18 @@
 # include <sys/mman.h>
 #endif
 
+/* On some platforms, this may not be defined.  On AIX, for example, this
+ * symbol is only defined when _NO_PROTO is defined, and _XOPEN_SOURCE is 500.
+ * How annoying.
+ */
+#ifndef MAP_FAILED
+# define MAP_FAILED	((void *) -1)
+#endif
+
+#if defined(PR_USE_CTRLS)
+# include <mod_ctrls.h>
+#endif /* PR_USE_CTRLS */
+
 /* Number of values to keep in a row. */
 #ifndef DELAY_NVALUES
 # define DELAY_NVALUES			256
@@ -55,6 +67,10 @@
 # define DELAY_SESS_NVALUES		16
 #endif
 
+#if defined(PR_USE_CTRLS)
+static ctrls_acttab_t delay_acttab[];
+#endif /* PR_USE_CTRLS */
+
 extern xaset_t *server_list;
 
 module delay_module;
@@ -188,8 +204,7 @@ static long delay_get_median(pool *p, unsigned int rownum, long interval) {
     *((long *) push_array(list)) = tab_vals[DELAY_NVALUES - i];
   *((long *) push_array(list)) = interval;
 
-  pr_log_debug(DEBUG6, MOD_DELAY_VERSION
-    ": selecting median interval from %d %s", list->nelts,
+  pr_trace_msg("delay", 6, "selecting median interval from %d %s", list->nelts,
     list->nelts != 1 ? "values" : "value");
   return delay_select_k(((list->nelts + 1) / 2), list);
 }
@@ -236,13 +251,13 @@ static void delay_delay(long interval) {
    */
   rand_usec = ((interval / 2.0) * rand()) / RAND_MAX;
   interval += rand_usec;
-  pr_log_debug(DEBUG10, MOD_DELAY_VERSION
-    ": additional random delay of %ld usecs added", (long int) rand_usec);
+  pr_trace_msg("delay", 8, "additional random delay of %ld usecs added",
+    (long int) rand_usec);
 
   tv.tv_sec = interval / 1000000;
   tv.tv_usec = interval % 1000000;
 
-  pr_log_debug(DEBUG0, MOD_DELAY_VERSION ": delaying for %ld usecs",
+  pr_trace_msg("delay", 8, "delaying for %ld usecs",
     (long int) ((tv.tv_sec * 1000000) + tv.tv_usec));
 
   delay_signals_block();
@@ -297,13 +312,14 @@ static int delay_table_init(void) {
     pr_log_debug(DEBUG0, MOD_DELAY_VERSION
       ": error opening DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
+    pr_trace_msg("delay", 1, "error opening DelayTable '%s': %s",
+      delay_tab.dt_path, strerror(errno));
     return -1;
   }
 
   if (pr_fsio_fstat(fh, &st) < 0) {
-    pr_log_debug(DEBUG0, MOD_DELAY_VERSION
-      ": error stat'ing DelayTable '%s': %s", delay_tab.dt_path,
-      strerror(errno));
+    pr_trace_msg("delay", 1, "error stat'ing DelayTable '%s': %s",
+      delay_tab.dt_path, strerror(errno));
     return -1;
   }
 
@@ -313,8 +329,8 @@ static int delay_table_init(void) {
      * configuration has changed by having vhosts added or removed.
      */
 
-    pr_log_debug(DEBUG2, MOD_DELAY_VERSION
-      ": expected table size %" PR_LU ", found %" PR_LU ", resetting table",
+    pr_trace_msg("delay", 3,
+      "expected table size %" PR_LU ", found %" PR_LU ", resetting table",
       (pr_off_t) tab_size, (pr_off_t) st.st_size);
     reset_table = TRUE;
   }
@@ -327,8 +343,7 @@ static int delay_table_init(void) {
     lock.l_start = 0;
     lock.l_len = 0;
 
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": write-locking DelayTable '%s'",
-      fh->fh_path);
+    pr_trace_msg("delay", 8, "write-locking DelayTable '%s'", fh->fh_path);
     while (fcntl(fh->fh_fd, F_SETLKW, &lock) < 0) {
       if (errno == EINTR) {
         pr_signals_handle();
@@ -338,6 +353,9 @@ static int delay_table_init(void) {
         pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
           "warning: unable to obtain write lock on DelayTable '%s': %s",
           fh->fh_path, strerror(errno));
+        pr_trace_msg("delay", 1,
+          "unable to obtain write lock on DelayTable '%s': %s", fh->fh_path,
+          strerror(errno));
         pr_fsio_close(fh);
         return -1;
       } 
@@ -349,23 +367,23 @@ static int delay_table_init(void) {
      * mmap() call.
      */
     lseek(fh->fh_fd, tab_size-1, SEEK_SET);
-    write(fh->fh_fd, "", 1);
+    (void) write(fh->fh_fd, "", 1);
 
     /* Truncate the table, in case we're shrinking an existing table. */
     pr_fsio_ftruncate(fh, tab_size);
 
     lock.l_type = F_UNLCK;
 
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": unlocking DelayTable '%s'",
-      fh->fh_path);
+    pr_trace_msg("delay", 8, "unlocking DelayTable '%s'", fh->fh_path);
     fcntl(fh->fh_fd, F_SETLK, &lock);
   }
 
   delay_tab.dt_fd = fh->fh_fd;
   delay_tab.dt_size = tab_size;
 
-  pr_log_debug(DEBUG10, MOD_DELAY_VERSION
-    ": mapping DelayTable '%s' into memory", fh->fh_path);
+  pr_trace_msg("delay", 8, "mapping DelayTable '%s' (%" PR_LU
+    " bytes, fd %d) into memory", fh->fh_path, (pr_off_t) delay_tab.dt_size,
+    delay_tab.dt_fd);
   delay_tab.dt_data = mmap(NULL, delay_tab.dt_size, PROT_READ|PROT_WRITE,
     MAP_SHARED, delay_tab.dt_fd, 0);
 
@@ -373,6 +391,8 @@ static int delay_table_init(void) {
     pr_log_pri(PR_LOG_ERR, MOD_DELAY_VERSION
       ": error mapping DelayTable '%s' into memory: %s", delay_tab.dt_path,
       strerror(errno));
+    pr_trace_msg("delay", 1, "error mapping DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(errno));
     pr_fsio_close(fh);
     return -1;
   }
@@ -417,8 +437,7 @@ static int delay_table_init(void) {
     lock.l_start = 0;
     lock.l_len = 0;
 
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": write-locking DelayTable '%s'",
-      fh->fh_path);
+    pr_trace_msg("delay", 8, "write-locking DelayTable '%s'", fh->fh_path);
     while (fcntl(fh->fh_fd, F_SETLKW, &lock) < 0) {
       if (errno == EINTR) {
         pr_signals_handle();
@@ -428,6 +447,9 @@ static int delay_table_init(void) {
         pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
           "warning: unable to obtain write lock on DelayTable '%s': %s",
           fh->fh_path, strerror(errno));
+        pr_trace_msg("delay", 1,
+          "unable to obtain write lock on DelayTable '%s': %s", fh->fh_path,
+          strerror(errno));
         pr_fsio_close(fh);
         return -1;
       }
@@ -439,25 +461,23 @@ static int delay_table_init(void) {
      * mmap() call.
      */
     lseek(fh->fh_fd, tab_size-1, SEEK_SET);
-    write(fh->fh_fd, "", 1);
+    (void) write(fh->fh_fd, "", 1);
 
     /* Truncate the table, in case we're shrinking an existing table. */
     pr_fsio_ftruncate(fh, tab_size);
 
-    pr_log_debug(DEBUG6, MOD_DELAY_VERSION ": resetting DelayTable '%s'",
-      delay_tab.dt_path);
+    pr_trace_msg("delay", 6, "resetting DelayTable '%s'", delay_tab.dt_path);
     delay_table_reset();
 
     lock.l_type = F_UNLCK;
 
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": unlocking DelayTable '%s'",
-      fh->fh_path);
+    pr_trace_msg("delay", 8, "unlocking DelayTable '%s'", fh->fh_path);
     fcntl(fh->fh_fd, F_SETLK, &lock);
   }
 
   /* Done */
-  pr_log_debug(DEBUG10, MOD_DELAY_VERSION
-    ": unmapping DelayTable '%s' from memory", delay_tab.dt_path);
+  pr_trace_msg("delay", 8, "unmapping DelayTable '%s' from memory",
+    delay_tab.dt_path);
   if (munmap(delay_tab.dt_data, delay_tab.dt_size) < 0) {
     int xerrno = errno;
     pr_fsio_close(fh);
@@ -466,6 +486,8 @@ static int delay_table_init(void) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
       ": error unmapping DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
+    pr_trace_msg("delay", 1, "error unmapping DelayTable '%s': %s",
+      delay_tab.dt_path, strerror(errno));
     return -1;
   }
 
@@ -475,6 +497,8 @@ static int delay_table_init(void) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
       ": error closing DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
+    pr_trace_msg("delay", 1, "error closing DelayTable '%s': %s",
+      delay_tab.dt_path, strerror(errno));
     return -1;
   }
 
@@ -491,7 +515,7 @@ static int delay_table_load(int lock_table) {
     lock.l_start = 0;
     lock.l_len = 0;
 
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": write-locking DelayTable '%s'",
+    pr_trace_msg("delay", 8, "write-locking DelayTable '%s'",
       delay_tab.dt_path);
     while (fcntl(delay_tab.dt_fd, F_SETLKW, &lock) < 0) {
       if (errno == EINTR) {
@@ -504,8 +528,9 @@ static int delay_table_load(int lock_table) {
   }
 
   if (!delay_tab.dt_data) {
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION
-      ": mapping DelayTable '%s' into memory", delay_tab.dt_path);
+    pr_trace_msg("delay", 8, "mapping DelayTable '%s' (%" PR_LU
+      " bytes, fd %d) into memory", delay_tab.dt_path,
+      (pr_off_t) delay_tab.dt_size, delay_tab.dt_fd);
     delay_tab.dt_data = mmap(NULL, delay_tab.dt_size, PROT_READ|PROT_WRITE,
       MAP_SHARED, delay_tab.dt_fd, 0);
 
@@ -550,16 +575,15 @@ static int delay_table_wlock(unsigned int rownum) {
   lock.l_start = sizeof(struct delay_rec) * rownum;
   lock.l_len = sizeof(struct delay_rec);
 
-  pr_log_debug(DEBUG10, MOD_DELAY_VERSION
-    ": write-locking DelayTable '%s', row %u", delay_tab.dt_path, rownum + 1);
+  pr_trace_msg("delay", 8, "write-locking DelayTable '%s', row %u",
+    delay_tab.dt_path, rownum + 1);
   while (fcntl(delay_tab.dt_fd, F_SETLKW, &lock) < 0) {
     if (errno == EINTR) {
       pr_signals_handle();
       continue;
     }
 
-    pr_log_debug(DEBUG2, MOD_DELAY_VERSION
-       ": error locking table: %s", strerror(errno));
+    pr_trace_msg("delay", 1, "error locking row: %s", strerror(errno));
     return -1;
   }
 
@@ -569,12 +593,14 @@ static int delay_table_wlock(unsigned int rownum) {
 static int delay_table_unload(int unlock_table) {
 
   if (delay_tab.dt_data) {
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION
-      ": unmapping DelayTable '%s' from memory", delay_tab.dt_path);
+    pr_trace_msg("delay", 8, "unmapping DelayTable '%s' from memory",
+      delay_tab.dt_path);
     if (munmap(delay_tab.dt_data, delay_tab.dt_size) < 0) {
       pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
         ": error unmapping DelayTable '%s': %s", delay_tab.dt_path,
         strerror(errno));
+      pr_trace_msg("delay", 1, "error unmapping DelayTable '%s': %s",
+        delay_tab.dt_path, strerror(errno));
       return -1;
     }
 
@@ -588,8 +614,7 @@ static int delay_table_unload(int unlock_table) {
     lock.l_start = 0;
     lock.l_len = 0;
 
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": unlocking DelayTable '%s'",
-      delay_tab.dt_path);
+    pr_trace_msg("delay", 8, "unlocking DelayTable '%s'", delay_tab.dt_path);
     while (fcntl(delay_tab.dt_fd, F_SETLK, &lock) < 0) {
       if (errno == EINTR) {
         pr_signals_handle();
@@ -611,25 +636,265 @@ static int delay_table_unlock(unsigned int rownum) {
   lock.l_start = sizeof(struct delay_rec) * rownum;
   lock.l_len = sizeof(struct delay_rec);
 
-  pr_log_debug(DEBUG10, MOD_DELAY_VERSION
-    ": unlocking DelayTable '%s', row %u", delay_tab.dt_path, rownum + 1);
+  pr_trace_msg("delay", 8, "unlocking DelayTable '%s', row %u",
+    delay_tab.dt_path, rownum + 1);
   while (fcntl(delay_tab.dt_fd, F_SETLKW, &lock) < 0) {
     if (errno == EINTR) {
       pr_signals_handle();
       continue;
     }
 
-    pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
-       ": error unlocking table: %s", strerror(errno));
+    pr_trace_msg("delay", 1, "error unlocking row: %s", strerror(errno));
     return -1;
   }
 
   return 0;
 }
 
+#if defined(PR_USE_CTRLS)
+
+/* Control handlers
+ */
+
+static int delay_handle_info(pr_ctrls_t *ctrl, int reqargc,
+    char **reqargv) {
+  register server_rec *s;
+  pool *tmp_pool;
+  pr_fh_t *fh;
+  char *vals;
+
+  PRIVS_ROOT
+  fh = pr_fsio_open(delay_tab.dt_path, O_RDWR);
+  PRIVS_RELINQUISH
+
+  if (!fh) {
+    pr_ctrls_add_response(ctrl,
+      "warning: unable to open DelayTable '%s': %s", delay_tab.dt_path,
+      strerror(errno));
+    return -1;
+  }
+
+  delay_tab.dt_fd = fh->fh_fd;
+  delay_tab.dt_data = NULL;
+
+  if (delay_table_load(TRUE) < 0) {
+    pr_ctrls_add_response(ctrl,
+      "unable to load DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(errno));
+    pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(errno));
+
+    pr_fsio_close(fh);
+    delay_tab.dt_fd = -1;
+    delay_tab.dt_data = NULL;
+
+    return -1;
+  }
+
+  tmp_pool = make_sub_pool(delay_pool);
+
+  for (s = (server_rec *) server_list->xas_list; s; s = s->next) {
+    unsigned int r = s->sid - 1;
+    register unsigned int i;
+ 
+    /* Row for USER values */
+    struct delay_rec *row = &((struct delay_rec *) delay_tab.dt_data)[r];
+    pr_ctrls_add_response(ctrl, "Address %s#%u: %u USER values (usecs):",
+      row->d_addr, row->d_port, row->d_nvals);
+
+    /* Start at the end of the row and work backward, as values are
+     * always added at the end of the row, shifting everything to the left.
+     */
+    vals = "";
+    for (i = 1; i < row->d_nvals; i++) {
+      char buf[80];
+
+      memset(buf, '\0', sizeof(buf));
+      sprintf(buf, "%10ld", row->d_vals[DELAY_NVALUES - i]);
+
+      vals = pstrcat(tmp_pool, vals, " ", buf, NULL);
+
+      if (i != 0 &&
+          i % 4 == 0) {
+        pr_ctrls_add_response(ctrl, "  %s", vals);
+        vals = "";
+      }
+    }
+
+    if (strlen(vals) > 0)
+      pr_ctrls_add_response(ctrl, "  %s", vals);
+
+    pr_ctrls_add_response(ctrl, "%s", "");
+
+    /* Row for PASS values */
+    row = &((struct delay_rec *) delay_tab.dt_data)[r + 1];
+    pr_ctrls_add_response(ctrl, "Address %s#%u: %u PASS values (usecs):",
+      row->d_addr, row->d_port, row->d_nvals);
+
+    vals = "";
+    for (i = 1; i < row->d_nvals; i++) {
+      char buf[80];
+
+      memset(buf, '\0', sizeof(buf));
+      sprintf(buf, "%10ld", row->d_vals[DELAY_NVALUES - i]);
+
+      vals = pstrcat(tmp_pool, vals, " ", buf, NULL);
+
+      if (i != 0 &&
+          i % 4 == 0) {
+        pr_ctrls_add_response(ctrl, "  %s", vals);
+        vals = "";
+      }
+    }
+
+    if (strlen(vals) > 0)
+      pr_ctrls_add_response(ctrl, "  %s", vals);
+  }
+  pr_ctrls_add_response(ctrl, "%s", "");
+
+  if (delay_table_unload(TRUE) < 0) {
+    pr_ctrls_add_response(ctrl,
+      "unable to unload DelayTable '%s' from memory: %s",
+      delay_tab.dt_path, strerror(errno));
+  }
+
+  pr_fsio_close(fh);
+
+  delay_tab.dt_fd = -1;
+  delay_tab.dt_data = NULL;
+
+  destroy_pool(tmp_pool);
+  return 0;
+}
+
+static int delay_handle_reset(pr_ctrls_t *ctrl, int reqargc,
+    char **reqarg) {
+  struct flock lock;
+  pr_fh_t *fh;
+
+  PRIVS_ROOT
+  fh = pr_fsio_open(delay_tab.dt_path, O_RDWR);
+  PRIVS_RELINQUISH
+
+  if (!fh) {
+    pr_ctrls_add_response(ctrl,
+      "unable to open DelayTable '%s': %s", delay_tab.dt_path,
+      strerror(errno));
+    return -1;
+  }
+
+  lock.l_type = F_WRLCK;
+  lock.l_whence = 0;
+  lock.l_start = 0;
+  lock.l_len = 0;
+
+  while (fcntl(fh->fh_fd, F_SETLKW, &lock) < 0) {
+    if (errno == EINTR) {
+      pr_signals_handle();
+      continue;
+
+    } else {
+      pr_ctrls_add_response(ctrl,
+        "unable to obtain write lock on DelayTable '%s': %s",
+         fh->fh_path, strerror(errno));
+      pr_fsio_close(fh);
+      return -1;
+    }
+  }
+
+  if (pr_fsio_ftruncate(fh, 0) < 0) {
+    pr_ctrls_add_response(ctrl,
+      "error truncating DelayTable '%s': %s", fh->fh_path, strerror(errno));
+    pr_fsio_close(fh);
+    return -1;
+  }
+
+  lock.l_type = F_UNLCK;
+  fcntl(fh->fh_fd, F_SETLK, &lock);
+
+  if (pr_fsio_close(fh) < 0) {
+    pr_ctrls_add_response(ctrl,
+      "error closing DelayTable '%s': %s", delay_tab.dt_path,
+      strerror(errno));
+    return -1;
+  }
+
+  pr_ctrls_add_response(ctrl, "DelayTable '%s' reset", delay_tab.dt_path);
+  return 0;
+}
+
+static int delay_handle_delay(pr_ctrls_t *ctrl, int reqargc,
+    char **reqargv) {
+
+  if (reqargc == 0 ||
+      reqargv == NULL) {
+    pr_ctrls_add_response(ctrl, "delay: missing required parameters");
+    return -1;
+  }
+
+  if (strcmp(reqargv[0], "info") == 0) {
+
+    if (!ctrls_check_acl(ctrl, delay_acttab, "info")) {
+      pr_ctrls_add_response(ctrl, "access denied");
+      return -1;
+    }
+
+    return delay_handle_info(ctrl, --reqargc, ++reqargv);
+
+  } else if (strcmp(reqargv[0], "reset") == 0) {
+
+    if (!ctrls_check_acl(ctrl, delay_acttab, "reset")) {
+      pr_ctrls_add_response(ctrl, "access denied");
+      return -1;
+    }
+
+    return delay_handle_reset(ctrl, --reqargc, ++reqargv);
+
+  } else {
+    pr_ctrls_add_response(ctrl, "unknown delay action: '%s'", reqargv[0]);
+    return -1;
+  }
+
+  return 0;
+}
+
+#endif /* PR_USE_CTRLS */
+
 /* Configuration handlers
  */
 
+/* usage: DelayControlsACLs actions|all allow|deny user|group list */
+MODRET set_delayctrlsacls(cmd_rec *cmd) {
+#if defined(PR_USE_CTRLS)
+  char *bad_action = NULL, **actions = NULL;
+
+  CHECK_ARGS(cmd, 4);
+  CHECK_CONF(cmd, CONF_ROOT);
+
+  actions = ctrls_parse_acl(cmd->tmp_pool, cmd->argv[1]);
+
+  /* Check the second parameter to make sure it is "allow" or "deny" */
+  if (strcmp(cmd->argv[2], "allow") != 0 &&
+      strcmp(cmd->argv[2], "deny") != 0)
+    CONF_ERROR(cmd, "second parameter must be 'allow' or 'deny'");
+
+  /* Check the third parameter to make sure it is "user" or "group" */
+  if (strcmp(cmd->argv[3], "user") != 0 &&
+      strcmp(cmd->argv[3], "group") != 0)
+    CONF_ERROR(cmd, "third parameter must be 'user' or 'group'");
+
+  bad_action = ctrls_set_module_acls(delay_acttab, delay_pool, actions,
+    cmd->argv[2], cmd->argv[3], cmd->argv[4]);
+  if (bad_action != NULL)
+    CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown delay action: '",
+      bad_action, "'", NULL));
+
+  return PR_HANDLED(cmd);
+#else
+  CONF_ERROR(cmd, "requires Controls support (--enable-ctrls)")
+#endif /* PR_USE_CTRLS */
+}
+
 /* usage: DelayEngine on|off */
 MODRET set_delayengine(cmd_rec *cmd) {
   config_rec *c;
@@ -646,7 +911,7 @@ MODRET set_delayengine(cmd_rec *cmd) {
   c->argv[0] = pcalloc(c->pool, sizeof(unsigned int));
   *((unsigned int *) c->argv[0]) = bool;
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: DelayTable path */
@@ -658,7 +923,7 @@ MODRET set_delaytable(cmd_rec *cmd) {
     CONF_ERROR(cmd, "must be an absolute path");
 
   add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* Command handlers
@@ -670,7 +935,7 @@ MODRET delay_post_pass(cmd_rec *cmd) {
   long interval, median;
 
   if (!delay_engine)
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
 
   /* We use sid-1, since the sid is a server number, and the locking
    * routines want a row index.  However, PASS rows are always after
@@ -684,7 +949,9 @@ MODRET delay_post_pass(cmd_rec *cmd) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
       "warning: unable to load DelayTable '%s' into memory: %s",
       delay_tab.dt_path, strerror(errno));
-    return DECLINED(cmd);
+    pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(errno));
+    return PR_DECLINED(cmd);
   }
 
   delay_table_wlock(rownum);
@@ -702,8 +969,7 @@ MODRET delay_post_pass(cmd_rec *cmd) {
    * poisoning the cache.
    */
   if (delay_npass < (DELAY_NVALUES / DELAY_SESS_NVALUES)) {
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": adding %ld usecs to PASS row",
-      interval);
+    pr_trace_msg("delay", 8, "adding %ld usecs to PASS row", interval);
     delay_table_add_interval(rownum, interval);
     delay_npass++;
 
@@ -727,15 +993,15 @@ MODRET delay_post_pass(cmd_rec *cmd) {
   if (interval < median)
     delay_delay(median - interval);
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET delay_pre_pass(cmd_rec *cmd) {
   if (!delay_engine)
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
 
   gettimeofday(&delay_tv, NULL);
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET delay_post_user(cmd_rec *cmd) {
@@ -744,7 +1010,7 @@ MODRET delay_post_user(cmd_rec *cmd) {
   long interval, median;
 
   if (!delay_engine)
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
 
   /* We use sid-1, since the sid is a server number, and the locking
    * routines want a row index.
@@ -756,7 +1022,9 @@ MODRET delay_post_user(cmd_rec *cmd) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
       "warning: unable to load DelayTable '%s' into memory: %s",
       delay_tab.dt_path, strerror(errno));
-    return DECLINED(cmd);
+    pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(errno));
+    return PR_DECLINED(cmd);
   }
 
   delay_table_wlock(rownum);
@@ -774,8 +1042,7 @@ MODRET delay_post_user(cmd_rec *cmd) {
    * poisoning the cache.
    */
   if (delay_nuser < (DELAY_NVALUES / DELAY_SESS_NVALUES)) {
-    pr_log_debug(DEBUG10, MOD_DELAY_VERSION ": adding %ld usecs to USER row",
-      interval);
+    pr_trace_msg("delay", 8, "adding %ld usecs to USER row", interval);
     delay_table_add_interval(rownum, interval);
     delay_nuser++;
 
@@ -799,15 +1066,15 @@ MODRET delay_post_user(cmd_rec *cmd) {
   if (interval < median)
     delay_delay(median - interval);
 
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 MODRET delay_pre_user(cmd_rec *cmd) {
   if (!delay_engine)
-    return DECLINED(cmd);
+    return PR_DECLINED(cmd);
 
   gettimeofday(&delay_tv, NULL);
-  return DECLINED(cmd);
+  return PR_DECLINED(cmd);
 }
 
 /* Event handlers
@@ -847,6 +1114,8 @@ static void delay_exit_ev(const void *event_data, void *user_data) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
       "warning: unable to load DelayTable '%s' into memory: %s",
       delay_tab.dt_path, strerror(errno));
+    pr_trace_msg("delay", 1, "unable to load DelayTable '%s' into memory: %s",
+      delay_tab.dt_path, strerror(errno));
     return;
   }
 
@@ -917,6 +1186,22 @@ static int delay_init(void) {
   delay_pool = make_sub_pool(permanent_pool);
   pr_pool_tag(delay_pool, MOD_DELAY_VERSION);
 
+#if defined(PR_USE_CTRLS)
+  if (pr_ctrls_register(&delay_module, "delay", "tune mod_delay settings",
+      delay_handle_delay) < 0) {
+    pr_log_pri(PR_LOG_INFO, MOD_DELAY_VERSION
+      ": error registering 'delay' control: %s", strerror(errno));
+
+  } else {
+    register unsigned int i;
+
+    for (i = 0; delay_acttab[i].act_action; i++) {
+      delay_acttab[i].act_acl = pcalloc(delay_pool, sizeof(ctrls_acl_t));
+      ctrls_init_acl(delay_acttab[i].act_acl);
+    }
+  }
+#endif /* PR_USE_CTRLS */
+
   return 0;
 }
 
@@ -942,8 +1227,7 @@ static int delay_sess_init(void) {
   delay_nuser = 0;
   delay_npass = 0;
 
-  pr_log_debug(DEBUG6, MOD_DELAY_VERSION ": opening DelayTable '%s'",
-    delay_tab.dt_path);
+  pr_trace_msg("delay", 6, "opening DelayTable '%s'", delay_tab.dt_path);
 
   PRIVS_ROOT
   fh = pr_fsio_open(delay_tab.dt_path, O_RDWR);
@@ -953,6 +1237,8 @@ static int delay_sess_init(void) {
     pr_log_pri(PR_LOG_WARNING, MOD_DELAY_VERSION
       ": warning: unable to open DelayTable '%s': %s", delay_tab.dt_path,
       strerror(errno));
+    pr_trace_msg("delay", 1, "unable to open DelayTable '%s': %s",
+      delay_tab.dt_path, strerror(errno));
     delay_engine = FALSE;
     return 0;
   }
@@ -966,7 +1252,16 @@ static int delay_sess_init(void) {
 /* Module API tables
  */
 
+#if defined(PR_USE_CTRLS)
+static ctrls_acttab_t delay_acttab[] = {
+  { "info",	NULL, NULL, NULL },
+  { "reset",	NULL, NULL, NULL },
+  { NULL,	NULL, NULL, NULL }
+};
+#endif /* PR_USE_CTRLS */
+
 static conftable delay_conftab[] = {
+  { "DelayControlsACLs",set_delayctrlsacls,	NULL },
   { "DelayEngine",	set_delayengine,	NULL },
   { "DelayTable",	set_delaytable,		NULL },
   { NULL }
