@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_dso -- support for loading/unloading modules at run-time
  *
- * Copyright (c) 2004 TJ Saunders <tj@castaglia.org>
+ * Copyright (c) 2004-2007 TJ Saunders <tj@castaglia.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,10 +22,10 @@
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
  *
- * This is mod_dso, contrib software for proftpd 1.2.x.
+ * This is mod_dso, contrib software for proftpd 1.3.x.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_dso.c,v 1.9 2004/12/17 21:56:00 castaglia Exp $
+ * $Id: mod_dso.c,v 1.13 2007/09/13 14:57:59 castaglia Exp $
  */
 
 #include "conf.h"
@@ -42,6 +42,8 @@ module dso_module;
 static const char *dso_module_path = PR_LIBEXEC_DIR;
 static pool *dso_pool = NULL;
 
+static const char *trace_channel = "dso";
+
 #ifdef PR_USE_CTRLS
 static ctrls_acttab_t dso_acttab[];
 #endif
@@ -53,6 +55,8 @@ static int dso_load_file(char *path) {
     return -1;
   }
 
+  pr_trace_msg(trace_channel, 5, "loading file '%s'", path);
+
   /* XXX Is this sufficient for loading an external library? */
   if (lt_dlopenext(path) == NULL) {
     pr_log_pri(PR_LOG_NOTICE, MOD_DSO_VERSION ": unable to open '%s': %s",
@@ -61,6 +65,7 @@ static int dso_load_file(char *path) {
     return -1;
   }
 
+  pr_trace_msg(trace_channel, 8, "file '%s' successfully loaded", path);
   return 0;
 }
 
@@ -92,6 +97,8 @@ static int dso_load_module(char *name) {
   /* Load file: $prefix/libexec/<module> */
   path = pdircat(dso_pool, dso_module_path, name, NULL);
 
+  pr_trace_msg(trace_channel, 5, "loading module '%s'", path);
+
   mh = lt_dlopenext(path);
   if (mh == NULL) {
     *tmp = '.';
@@ -124,6 +131,9 @@ static int dso_load_module(char *name) {
     pr_log_debug(DEBUG1, MOD_DSO_VERSION
       ": unable to find module symbol '%s' in '%s'", symbol_name,
         mh ? name : "self");
+    pr_trace_msg(trace_channel, 1, "unable to find module symbol '%s' in '%s'",
+      symbol_name, mh ? name : "self");
+
     lt_dlclose(mh);
     mh = NULL;
 
@@ -137,24 +147,33 @@ static int dso_load_module(char *name) {
   /* Add the module to the core structures */
   res = pr_module_load(m);
   if (res < 0) {
-    if (errno == EEXIST)
+    if (errno == EEXIST) {
       pr_log_pri(PR_LOG_INFO, MOD_DSO_VERSION
         ": module 'mod_%s.c' already loaded", m->name);
+      pr_trace_msg(trace_channel, 1, "module 'mod_%s.c' already loaded",
+        m->name);
 
-    else if (errno == EACCES)
+    } else if (errno == EACCES) {
       pr_log_pri(PR_LOG_ERR, MOD_DSO_VERSION
         ": module 'mod_%s.c' has wrong API version (0x%x), must be 0x%x",
         m->name, m->api_version, PR_MODULE_API_VERSION);
+      pr_trace_msg(trace_channel, 1,
+        "module 'mod_%s.c' has wrong API version (0x%x), must be 0x%x",
+        m->name, m->api_version, PR_MODULE_API_VERSION);
 
-    else if (errno == EPERM)
+    } else if (errno == EPERM) {
       pr_log_pri(PR_LOG_ERR, MOD_DSO_VERSION
         ": module 'mod_%s.c' failed to initialize", m->name);
+      pr_trace_msg(trace_channel, 1, "module 'mod_%s.c' failed to initialize",
+        m->name);
+    }
 
     lt_dlclose(mh);
     mh = NULL;
     return -1;
   }
-  
+
+  pr_trace_msg(trace_channel, 8, "module '%s' successfully loaded", path);
   return 0;
 }
 
@@ -170,10 +189,15 @@ static int dso_unload_module(module *m) {
 
   name = pstrdup(dso_pool, m->name);
 
+  pr_trace_msg(trace_channel, 5, "unloading module 'mod_%s.c'", name);
+
   res = pr_module_unload(m);
-  if (res < 0)
+  if (res < 0) {
     pr_log_debug(DEBUG1, MOD_DSO_VERSION
       ": error unloading module 'mod_%s.c': %s", m->name, strerror(errno));
+    pr_trace_msg(trace_channel, 1,
+      "error unloading module 'mod_%s.c': %s", m->name, strerror(errno));
+  }
 
   if (lt_dlclose(m->handle) < 0) {
     pr_log_debug(DEBUG1, MOD_DSO_VERSION ": error closing '%s': %s",
@@ -181,6 +205,8 @@ static int dso_unload_module(module *m) {
     return -1;
   }
 
+  pr_trace_msg(trace_channel, 8, "module 'mod_%s.c' successfully unloaded",
+    name);
   return 0;
 }
 
@@ -348,7 +374,7 @@ MODRET set_loadfile(cmd_rec *cmd) {
     CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "error loading '", cmd->argv[1],
       "': ", strerror(errno), NULL));
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: LoadModule module */
@@ -361,7 +387,7 @@ MODRET set_loadmodule(cmd_rec *cmd) {
     CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, "error loading module '",
       cmd->argv[1], "': ", strerror(errno), NULL));
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: ModuleControlsACLs actions|all allow|deny user|group list */
@@ -382,12 +408,13 @@ MODRET set_modulectrlsacls(cmd_rec *cmd) {
       strcmp(cmd->argv[3], "group") != 0)
     CONF_ERROR(cmd, "third parameter must be 'user' or 'group'");
 
-  if ((bad_action = ctrls_set_module_acls(dso_acttab, dso_pool, actions,
-      cmd->argv[2], cmd->argv[3], cmd->argv[4])) != NULL)
+  bad_action = ctrls_set_module_acls(dso_acttab, dso_pool, actions,
+    cmd->argv[2], cmd->argv[3], cmd->argv[4]);
+  if (bad_action != NULL)
     CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown action: '",
       bad_action, "'", NULL));
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 #else
   CONF_ERROR(cmd, "requires Controls support (--enable-ctrls)");
 #endif
@@ -477,7 +504,7 @@ MODRET set_moduleorder(cmd_rec *cmd) {
   for (m = loaded_modules; m; m = m->next)
     pr_log_pri(PR_LOG_NOTICE, " mod_%s.c", m->name);
 
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* usage: ModulePath path */
@@ -502,7 +529,7 @@ MODRET set_modulepath(cmd_rec *cmd) {
       NULL));
 
   if (st.st_mode & S_IWOTH)
-    CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, cmd->argv[1], " is world-writeable",
+    CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, cmd->argv[1], " is world-writable",
       NULL));
 
   if (lt_dlsetsearchpath(cmd->argv[1]) < 0)
@@ -510,7 +537,7 @@ MODRET set_modulepath(cmd_rec *cmd) {
       lt_dlerror(), NULL));
 
   dso_module_path = pstrdup(dso_pool, cmd->argv[1]);
-  return HANDLED(cmd);
+  return PR_HANDLED(cmd);
 }
 
 /* Event handlers
@@ -542,7 +569,9 @@ static void dso_restart_ev(const void *event_data, void *user_data) {
 
   /* Unload all shared modules. */
   for (mi = loaded_modules; mi; mi = m) {
+#ifndef PR_USE_CTRLS
     register unsigned int i;
+#endif /* PR_USE_CTRLS */
     int is_static = FALSE;
 
     m = mi->next;
