@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2006 The ProFTPD Project team
  *  
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,12 +24,12 @@
 
 /* Controls API routines
  *
- * $Id: ctrls.c,v 1.9 2004/05/26 19:25:42 castaglia Exp $
+ * $Id: ctrls.c,v 1.16 2006/12/16 01:13:52 castaglia Exp $
  */
 
 #include "conf.h"
 
-#ifdef USE_CTRLS
+#ifdef PR_USE_CTRLS
 
 typedef struct ctrls_act_obj {
   struct ctrls_act_obj *prev, *next;
@@ -50,6 +50,10 @@ static ctrls_action_t *ctrls_action_list = NULL;
 static pr_ctrls_t *ctrls_active_list = NULL;
 static pr_ctrls_t *ctrls_free_list = NULL;
 
+static int ctrls_use_isfifo = FALSE;
+
+static const char *trace_channel = "ctrls";
+
 /* lookup/lookup_next indices */
 static ctrls_action_t *action_lookup_next = NULL;
 static const char *action_lookup_action = NULL;
@@ -222,6 +226,10 @@ int pr_ctrls_register(const module *mod, const char *action,
     return -1;
   }
 
+  pr_trace_msg("ctrls", 3,
+    "module '%s' registering handler for ctrl action '%s' (at %p)",
+    mod ? mod->name : "(none)", action, cb);
+
   /* Block ctrls while we're doing this */
   pr_block_ctrls();
 
@@ -276,6 +284,10 @@ int pr_ctrls_unregister(module *mod, const char *action) {
     return -1;
   }
 
+  pr_trace_msg("ctrls", 3,
+    "module '%s' unregistering handler for ctrl action '%s'",
+    mod ? mod->name : "(none)", action);
+
   /* Make sure that ctrls are blocked while we're doing this */
   pr_block_ctrls();
 
@@ -524,11 +536,20 @@ int pr_ctrls_recv_request(pr_ctrls_cl_t *cl) {
     return -1;
   }
 
+  if (reqarglen >= sizeof(reqaction)) {
+    pr_signals_unblock();
+    errno = ENOMEM;
+    return -1;
+  }
+
+  memset(reqaction, '\0', sizeof(reqaction));
+
   if (read(cl->cl_fd, reqaction, reqarglen) < 0) {
     pr_signals_unblock();
     return -1;
   }
 
+  reqaction[sizeof(reqaction)-1] = '\0';
   nreqargs--;
 
   /* Find a matching action object, and use it to populate a ctrl object,
@@ -647,17 +668,16 @@ int pr_ctrls_recv_response(pool *resp_pool, int ctrls_sockfd,
       return -1;
     }
 
-    memset(response, '\0', sizeof(response));
-
     /* Make sure resparglen is not too big */
-    if (resparglen > sizeof(response)) {
+    if (resparglen >= sizeof(response)) {
       pr_signals_unblock();
       errno = ENOMEM;
       return -1;
     }
 
-    bread = read(ctrls_sockfd, response, resparglen);
+    memset(response, '\0', sizeof(response));
 
+    bread = read(ctrls_sockfd, response, resparglen);
     while (bread != resparglen) {
       if (bread < 0) {
         pr_signals_unblock(); 
@@ -931,6 +951,26 @@ int pr_ctrls_connect(const char *socket_file) {
   return sockfd;
 }
 
+int pr_ctrls_issock_unix(mode_t sock_mode) {
+
+  if (ctrls_use_isfifo) {
+#ifdef S_ISFIFO
+    if (S_ISFIFO(sock_mode)) {
+      return 0;
+    }
+#endif /* S_ISFIFO */
+  } else {
+#ifdef S_ISSOCK
+    if (S_ISSOCK(sock_mode)) {
+      return 0;
+    }
+#endif /* S_ISSOCK */
+  }
+
+  errno = ENOSYS;
+  return -1;
+}
+
 void pr_block_ctrls(void) {
   ctrls_blocked = TRUE;
 }
@@ -991,7 +1031,8 @@ int pr_run_ctrls(module *mod, const char *action) {
 
     if (action &&
         strcmp(ctrl->ctrls_action, action) == 0) {
-      pr_log_debug(DEBUG7, "calling '%s' control handler", ctrl->ctrls_action);
+      pr_trace_msg(trace_channel, 7, "calling '%s' control handler",
+        ctrl->ctrls_action);
 
       /* Invoke the callback, if the ctrl's action matches.  Unblock
        * ctrls before invoking the callback, then re-block them after the
@@ -1011,7 +1052,8 @@ int pr_run_ctrls(module *mod, const char *action) {
       }
 
     } else if (!action) {
-      pr_log_debug(DEBUG5, "calling '%s' control handler", ctrl->ctrls_action);
+      pr_trace_msg(trace_channel, 7, "calling '%s' control handler",
+        ctrl->ctrls_action);
 
       /* If no action was given, invoke every callback */
       pr_unblock_ctrls();
@@ -1061,8 +1103,15 @@ int pr_reset_ctrls(void) {
 }
 
 void init_ctrls(void) {
-  if (ctrls_pool)
+  struct stat st;
+  int sockfd;
+  struct sockaddr_un sockun;
+  size_t socklen;
+  char *sockpath = PR_RUN_DIR "/test.sock";
+
+  if (ctrls_pool) {
     destroy_pool(ctrls_pool);
+  }
 
   ctrls_pool = make_sub_pool(permanent_pool);
   pr_pool_tag(ctrls_pool, "Controls Pool");
@@ -1077,7 +1126,67 @@ void init_ctrls(void) {
   action_lookup_action = NULL;
   action_lookup_module = NULL;
 
+  /* Run-time check to find out whether this platform identifies a
+   * Unix domain socket file descriptor via the S_ISFIFO macro, or
+   * the S_ISSOCK macro.
+   */
+
+  sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
+  if (sockfd < 0) {
+    pr_log_pri(PR_LOG_NOTICE, "notice: unable to create Unix domain socket: %s",
+      strerror(errno));
+    return;
+  }
+
+  memset(&sockun, 0, sizeof(sockun));
+  sockun.sun_family = AF_UNIX;
+  sstrncpy(sockun.sun_path, sockpath, strlen(sockpath) + 1);
+  socklen = sizeof(struct sockaddr_un);
+
+  if (bind(sockfd, (struct sockaddr *) &sockun, socklen) < 0) {
+    pr_log_pri(PR_LOG_NOTICE,
+      "notice: unable to bind to Unix domain socket at '%s': %s",
+      sockpath, strerror(errno));
+    (void) close(sockfd);
+    (void) unlink(sockpath);
+    return;
+  }
+
+  if (fstat(sockfd, &st) < 0) {
+    pr_log_pri(PR_LOG_NOTICE,
+      "notice: unable to stat Unix domain socket at '%s': %s",
+      sockpath, strerror(errno));
+    (void) close(sockfd);
+    (void) unlink(sockpath);
+    return;
+  }
+
+#ifdef S_ISFIFO
+  pr_trace_msg(trace_channel, 9, "testing Unix domain socket using S_ISFIFO");
+  if (S_ISFIFO(st.st_mode)) {
+    ctrls_use_isfifo = TRUE;
+  }
+#else
+  pr_trace_msg(trace_channel, 9, "cannot test Unix domain socket using "
+    "S_ISFIFO: macro undefined");
+#endif
+
+#ifdef S_ISSOCK
+  pr_trace_msg(trace_channel, 9, "testing Unix domain socket using S_ISSOCK");
+  if (S_ISSOCK(st.st_mode)) {
+    ctrls_use_isfifo = FALSE;
+  }
+#else
+  pr_trace_msg(trace_channel, 9, "cannot test Unix domain socket using "
+    "S_ISSOCK: macro undefined");
+#endif
+
+  pr_trace_msg(trace_channel, 9, "using %s macro for Unix domain socket "
+    "detection", ctrls_use_isfifo ? "S_ISFIFO" : "S_ISSOCK");
+
+  (void) close(sockfd);
+  (void) unlink(sockpath);
   return;
 }
 
-#endif /* USE_CTRLS */
+#endif /* PR_USE_CTRLS */
