@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2004 The ProFTPD Project team
+ * Copyright (c) 2001-2007 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,9 +24,8 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/*
- * Data connection management functions
- * $Id: data.c,v 1.89 2006/02/08 01:25:46 castaglia Exp $
+/* Data connection management functions
+ * $Id: data.c,v 1.94 2007/05/15 00:41:19 castaglia Exp $
  */
 
 #include "conf.h"
@@ -41,6 +40,8 @@
 #include <sys/uio.h>
 #endif /* HAVE_SYS_UIO_H */
 
+static const char *trace_channel = "data";
+
 /* local macro */
 
 #define MODE_STRING	(session.sf_flags & (SF_ASCII|SF_ASCII_OVERRIDE) ? \
@@ -66,14 +67,16 @@ static int stalled_timeout_cb(CALLBACK_FRAME) {
   return 0;
 }
 
-/* this signal is raised if we get OOB data on the control connection, and
- * a data transfer is in progress
+/* This signal is raised if we get OOB data on the control connection, and
+ * a data transfer is in progress.
  */
-static RETSIGTYPE data_urgent(int sig) {
-  if (nstrm) {
-    session.sf_flags |= SF_ABORT;
+static RETSIGTYPE data_urgent(int signo) {
+  pr_trace_msg(trace_channel, 5, "received SIGURG signal (signal %d), "
+    "setting 'aborted' session flag", signo);
+  session.sf_flags |= SF_ABORT;
+
+  if (nstrm)
     pr_netio_abort(nstrm);
-  }
 
   signal(SIGURG, data_urgent);
 }
@@ -278,10 +281,10 @@ static int data_pasv_open(char *reason, off_t size) {
 
     if (session.xfer.xfer_type != STOR_UNIQUE) {
       if (size)
-        pr_response_send(R_150, "Opening %s mode data connection for %s "
-          "(%" PR_LU " bytes)", MODE_STRING, reason, (pr_off_t) size);
+        pr_response_send(R_150, _("Opening %s mode data connection for %s "
+          "(%" PR_LU " bytes)"), MODE_STRING, reason, (pr_off_t) size);
       else
-        pr_response_send(R_150, "Opening %s mode data connection for %s",
+        pr_response_send(R_150, _("Opening %s mode data connection for %s"),
           MODE_STRING, reason);
 
     } else {
@@ -316,7 +319,7 @@ static int data_pasv_open(char *reason, off_t size) {
     pr_log_pri(PR_LOG_ERR, "Error: unable to accept an incoming data "
       "connection (%s)", strerror(c->xerrno));
 
-  pr_response_add_err(R_425, "Unable to build data connection: %s",
+  pr_response_add_err(R_425, _("Unable to build data connection: %s"),
     strerror(session.d->xerrno));
   destroy_pool(session.d->pool);
   session.d = NULL;
@@ -358,7 +361,7 @@ static int data_active_open(char *reason, off_t size) {
 
   if (pr_inet_connect(session.d->pool, session.d, &session.data_addr,
       session.data_port) == -1) {
-    pr_response_add_err(R_425, "Unable to build data connection: %s",
+    pr_response_add_err(R_425, _("Unable to build data connection: %s"),
       strerror(session.d->xerrno));
     destroy_pool(session.d->pool);
     session.d = NULL;
@@ -378,10 +381,10 @@ static int data_active_open(char *reason, off_t size) {
 
     if (session.xfer.xfer_type != STOR_UNIQUE) {
       if (size)
-        pr_response_send(R_150, "Opening %s mode data connection for %s "
-          "(%" PR_LU " bytes)", MODE_STRING, reason, (pr_off_t) size);
+        pr_response_send(R_150, _("Opening %s mode data connection for %s "
+          "(%" PR_LU " bytes)"), MODE_STRING, reason, (pr_off_t) size);
       else
-        pr_response_send(R_150, "Opening %s mode data connection for %s",
+        pr_response_send(R_150, _("Opening %s mode data connection for %s"),
           MODE_STRING, reason);
 
     } else {
@@ -414,7 +417,7 @@ static int data_active_open(char *reason, off_t size) {
     return 0;
   }
 
-  pr_response_add_err(R_425, "Unable to build data connection: %s",
+  pr_response_add_err(R_425, _("Unable to build data connection: %s"),
     strerror(session.d->xerrno));
   destroy_pool(session.d->pool);
   session.d = NULL;
@@ -511,10 +514,8 @@ int pr_data_open(char *filename, char *reason, int direction, off_t size) {
     if (TimeoutNoXfer)
       pr_timer_reset(TIMER_NOXFER, ANY_MODULE);
 
-    /* Allow aborts. */
-
-    /* Set the current NetIO stream to allow interrupted syscalls, so our
-     * SIGURG handler can interrupt it
+    /* Allow aborts -- set the current NetIO stream to allow interrupted
+     * syscalls, so our SIGURG handler can interrupt it
      */
     pr_netio_set_poll_interval(nstrm, 1);
 
@@ -531,12 +532,19 @@ int pr_data_open(char *filename, char *reason, int direction, off_t size) {
 #ifdef SA_INTERRUPT
     act.sa_flags |= SA_INTERRUPT;
 #endif
-    sigaction(SIGURG, &act, NULL);
+
+    if (sigaction(SIGURG, &act, NULL) < 0)
+      pr_log_pri(PR_LOG_WARNING,
+        "warning: unable to set SIGURG signal handler: %s", strerror(errno));
+
 #ifdef HAVE_SIGINTERRUPT
     /* This is the BSD way of ensuring interruption.
      * Linux uses it too (??)
      */
-    siginterrupt(SIGURG, 1);
+    if (siginterrupt(SIGURG, 1) < 0)
+      pr_log_pri(PR_LOG_WARNING,
+        "warning: unable to make SIGURG interrupt system calls: %s",
+        strerror(errno));
 #endif
   }
 
@@ -566,7 +574,7 @@ void pr_data_close(int quiet) {
   session_set_idle();
 
   if (!quiet)
-    pr_response_add(R_226, "Transfer complete.");
+    pr_response_add(R_226, _("Transfer complete"));
 }
 
 /* Note: true_abort may be false in real abort situations, because
@@ -588,7 +596,7 @@ void pr_data_cleanup(void) {
   if (session.xfer.p)
     destroy_pool(session.xfer.p);
 
-  memset(&session.xfer,0,sizeof(session.xfer));
+  memset(&session.xfer, 0, sizeof(session.xfer));
 }
 
 /* In order to avoid clearing the transfer counters in session.xfer, we don't
@@ -627,7 +635,6 @@ void pr_data_abort(int err, int quiet) {
 
   if (!quiet) {
     char	*respcode = R_426;
-    char	*fmt = NULL;
     char	*msg = NULL;
     char	msgbuf[64];
 
@@ -635,13 +642,13 @@ void pr_data_abort(int err, int quiet) {
 
     case 0:
       respcode = R_426;
-      msg = "Data connection closed.";
+      msg = _("Data connection closed");
       break;
 
 #ifdef ENXIO
     case ENXIO:
       respcode = R_451;
-      msg = "Unexpected streams hangup.";
+      msg = _("Unexpected streams hangup");
       break;
 
 #endif
@@ -654,7 +661,7 @@ void pr_data_abort(int err, int quiet) {
 #endif
 #if defined(EAGAIN) || defined(ENOMEM)
       respcode = R_451;
-      msg = "Insufficient memory or file locked.";
+      msg = _("Insufficient memory or file locked");
       break;
 #endif
 
@@ -760,7 +767,7 @@ void pr_data_abort(int err, int quiet) {
 	|| defined(ENOLINK) || defined(ENOLCK) || defined(ENETRESET) \
 	|| defined(ECONNABORTED) || defined(ECONNRESET) || defined(ETIMEDOUT)
       respcode = R_450;
-      msg = "Link to file server lost.";
+      msg = _("Link to file server lost");
       break;
 #endif
     }
@@ -769,7 +776,7 @@ void pr_data_abort(int err, int quiet) {
         (msg = strerror(err)) == NULL ) {
 
       if (snprintf(msgbuf, sizeof(msgbuf),
-          "Unknown or out of range errno [%d]", err) > 0)
+          _("Unknown or out of range errno [%d]"), err) > 0)
 	msg = msgbuf;
     }
 
@@ -780,8 +787,7 @@ void pr_data_abort(int err, int quiet) {
      * and we don't want to add another to the error queue.
      */
     if (!true_abort)
-      pr_response_add_err(respcode, fmt ? fmt : "Transfer aborted. %s",
-        msg ? msg : "");
+      pr_response_add_err(respcode, _("Transfer aborted. %s"), msg ? msg : "");
   }
 
   if (true_abort)
@@ -1004,6 +1010,10 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
       /* Under Linux semantics, this occurs when a signal has interrupted
        * sendfile().
        */
+      if (XFER_ABORTED) {
+        errno = EINTR;
+        return -1;
+      }
 
       count -= len;
 
@@ -1092,6 +1102,9 @@ pr_sendfile_t pr_data_sendfile(int retr_fd, off_t *offset, off_t count) {
        * For obvious reasons, HP/UX sendfile is not supported yet.
        */
       if (errno == EINTR) {
+        if (XFER_ABORTED)
+          return -1;
+
         pr_signals_handle();
 
         /* If we got everything in this transaction, we're done. */
