@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003, 2004 The ProFTPD Project team
+ * Copyright (c) 2001-2007 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* Read configuration file(s), and manage server/configuration structures.
- * $Id: dirtree.c,v 1.171 2005/09/28 02:06:26 castaglia Exp $
+ * $Id: dirtree.c,v 1.185 2007/07/24 22:09:39 castaglia Exp $
  */
 
 #include "conf.h"
@@ -54,16 +54,19 @@ char MultilineRFC2228 = 0;
 extern pool *global_config_pool;
 
 /* Used by find_config_* */
-xaset_t *find_config_top = NULL;
+static xaset_t *find_config_top = NULL;
 
-static void _mergedown(xaset_t *, int);
+static void merge_down(xaset_t *, int);
 
-/* Used by get_param_int_next & get_param_ptr_next as "placeholders" */
-static config_rec *_last_param_int = NULL;
 static config_rec *_last_param_ptr = NULL;
 static unsigned char _kludge_disable_umask = 0;
 
-array_header *server_defines = NULL;
+static pool *defines_pool = NULL;
+static array_header *defines_list = NULL;
+
+static pool *config_tab_pool = NULL;
+static pr_table_t *config_tab = NULL;
+static unsigned int config_id = 0;
 
 static int allow_dyn_config(void) {
   config_rec *c = NULL;
@@ -261,8 +264,8 @@ char *path_subst_uservar(pool *path_pool, char **path) {
    */
 
   substr_path = *path;
-
-  while ((substr = strstr(substr_path, "%u[")) != NULL) {
+  substr = substr_path ? strstr(substr_path, "%u[") : NULL;
+  while (substr) {
     int i = 0;
     char *substr_end = NULL, *substr_dup = NULL, *endp = NULL;
     char ref_char[2] = {'\0', '\0'};
@@ -270,10 +273,11 @@ char *path_subst_uservar(pool *path_pool, char **path) {
     /* Now, find the closing ']'. If not found, it is a syntax error;
      * continue on without processing this occurrence.
      */
-    if ((substr_end = strchr(substr, ']')) == NULL)
-
+    substr_end = strchr(substr, ']');
+    if (substr_end == NULL) {
       /* Just end here. */
       break;
+    }
 
     /* Make a copy of the entire substring. */
     substr_dup = pstrdup(path_pool, substr);
@@ -335,11 +339,14 @@ char *path_subst_uservar(pool *path_pool, char **path) {
      * referenced character/string.
      */
     substr_path = sreplace(path_pool, substr_path, substr_dup, ref_char, NULL);
+    substr = substr_path ? strstr(substr_path, "%u[") : NULL;
   }
 
   /* Check for any bare "%u", and handle those if present. */
-  if (strstr(substr_path, "%u"))
+  if (substr_path &&
+      strstr(substr_path, "%u"))
     new_path = sreplace(path_pool, substr_path, "%u", session.user, NULL);
+
   else
     new_path = substr_path;
 
@@ -487,21 +494,44 @@ unsigned char dir_hide_file(const char *path) {
   return FALSE;	
 }
 
-unsigned char define_exists(const char *definition) {
+static void define_restart_ev(const void *event_data, void *user_data) {
+  if (defines_pool) {
+    destroy_pool(defines_pool);
+    defines_pool = NULL;
+    defines_list = NULL;
+  }
 
-  /* Check the list of specified definitions, if present.
-   */
-  if (server_defines) {
-    char **defines = server_defines->elts;
+  pr_event_unregister(NULL, "core.restart", define_restart_ev);
+}
+
+int pr_define_add(const char *definition) {
+
+  if (!defines_list) {
+    defines_pool = make_sub_pool(permanent_pool);
+    pr_pool_tag(defines_pool, "Defines Pool");
+    defines_list = make_array(defines_pool, 0, sizeof(char *));
+
+    pr_event_register(NULL, "core.restart", define_restart_ev, NULL);
+  }
+
+  *((char **) push_array(defines_list)) = pstrdup(defines_pool, definition);
+  return 0;
+}
+
+unsigned char pr_define_exists(const char *definition) {
+
+  if (defines_list) {
+    char **defines = defines_list->elts;
     register unsigned int i = 0;
 
-    for (i = 0; i < server_defines->nelts; i++) {
-      if (defines[i] && !strcmp(defines[i], definition))
+    for (i = 0; i < defines_list->nelts; i++) {
+      if (defines[i] &&
+          strcmp(defines[i], definition) == 0)
         return TRUE;
     }
   }
 
-  /* default */
+  errno = ENOENT;
   return FALSE;
 }
 
@@ -603,7 +633,7 @@ config_rec *add_config_set(xaset_t **set, const char *name) {
     set_pool = make_sub_pool(permanent_pool);
     pr_pool_tag(set_pool, "config set pool");
 
-    *set = xaset_create(set_pool,NULL);
+    *set = xaset_create(set_pool, NULL);
     (*set)->pool = set_pool;
 
     /* Now, make a subpool for the config_rec to be allocated. */
@@ -626,9 +656,13 @@ config_rec *add_config_set(xaset_t **set, const char *name) {
   c->pool = conf_pool;
   c->set = *set;
   c->parent = parent;
-  if (name)
+
+  if (name) {
     c->name = pstrdup(conf_pool, name);
-  xaset_insert_end(*set, (xasetmember_t*)c);
+    c->config_id = pr_config_set_id(name);
+  }
+
+  xaset_insert_end(*set, (xasetmember_t *) c);
 
   return c;
 }
@@ -880,8 +914,8 @@ unsigned char pr_expr_eval_user_or(char **expr) {
 }
 /* Per-directory configuration */
 
-static int _strmatch(register char *s1, register char *s2) {
-  register int len = 0;
+static size_t _strmatch(register char *s1, register char *s2) {
+  register size_t len = 0;
 
   while (*s1 && *s2 && *s1++ == *s2++)
     len++;
@@ -925,20 +959,7 @@ static config_rec *recur_match_path(pool *p, xaset_t *s, char *path) {
         tmp_path = pdircat(p, tmp_path, "*", NULL);
       }
 
-      /* Temporary measure until we figure what's going on with
-       * gnu fnmatch
-       *
-       * Hmm...wonder what this is, and if it's still an issue.  I love
-       * cryptic comments in other people's code. :)
-       *
-       * - MacGyver
-       */
-
-#if 0
-      if (pr_fnmatch(tmp_path, path, PR_FNM_PATHNAME) == 0) {
-#else
       if (pr_fnmatch(tmp_path, path, 0) == 0) {
-#endif
         if (c->subset) {
           if ((res = recur_match_path(p, c->subset, path)))
             return res;
@@ -1588,18 +1609,22 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
 
   memcpy(&st, stp, sizeof(st));
 
-  if (S_ISDIR(st.st_mode))
+  if (S_ISDIR(st.st_mode)) {
     dynpath = pdircat(p, path, "/.ftpaccess", NULL);
-  else
+
+  } else
     dynpath = NULL;
 
   while (path) {
+    pr_signals_handle();
+
     if (session.chroot_path) {
       fullpath = pdircat(p, session.chroot_path, path, NULL);
 
       if (strcmp(fullpath, "/") &&
-          *(fullpath + strlen(fullpath) - 1) == '/')
+          *(fullpath + strlen(fullpath) - 1) == '/') {
         *(fullpath + strlen(fullpath) - 1) = '\0';
+      }
 
     } else
       fullpath = path;
@@ -1612,9 +1637,10 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
 
     d = dir_match_path(p, fullpath);
 
-    if (!d && isfile != -1) {
+    if (!d &&
+        isfile != -1) {
       set = (session.anon_config ? &session.anon_config->subset :
-             &main_server->conf);
+        &main_server->conf);
 
       d = add_config_set(set, fullpath);
       d->config_type = CONF_DIR;
@@ -1622,7 +1648,7 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
       d->argv = pcalloc(d->pool, 2 * sizeof (void *));
 
     } else if (d) {
-      config_rec *newd,*dnext;
+      config_rec *newd, *dnext;
 
       if (isfile != -1 &&
           strcmp(d->name, fullpath) != 0) {
@@ -1641,10 +1667,11 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
 
         set = (d->parent ? &d->parent->subset : &main_server->conf);
 
-	if (d->subset && d->subset->xas_list) {
+	if (d->subset &&
+            d->subset->xas_list) {
 
        	  /* Remove all old dynamic entries. */
-          for (newd = (config_rec *)d->subset->xas_list; newd; newd = dnext) {
+          for (newd = (config_rec *) d->subset->xas_list; newd; newd = dnext) {
 	    dnext = newd->next;
 
             if (newd->flags & CF_DYNAMIC) {
@@ -1654,7 +1681,8 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
           }
 	}
 
-        if (d->subset && !d->subset->xas_list) {
+        if (d->subset &&
+            !d->subset->xas_list) {
           destroy_pool(d->subset->pool);
           d->subset = NULL;
           d->argv[0] = NULL;
@@ -1668,7 +1696,9 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
       }
     }
 
-    if (isfile != -1 && d && st.st_size > 0 &&
+    if (isfile != -1 &&
+        d &&
+        st.st_size > 0 &&
         st.st_mtime > (d->argv[0] ? *((time_t *) d->argv[0]) : 0)) {
 
       /* File has been modified or not loaded yet */
@@ -1686,12 +1716,15 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
       d->config_type = CONF_DIR;
       pr_parser_cleanup();
 
-      _mergedown(*set, TRUE);
+      merge_down(*set, TRUE);
     }
 
-    if (isfile == -1 && removed && d && set) {
+    if (isfile == -1 &&
+        removed &&
+        d &&
+        set) {
       pr_log_debug(DEBUG5, "dynamic configuration removed for %s", fullpath);
-      _mergedown(*set, FALSE);
+      merge_down(*set, FALSE);
     }
 
     if (!recurse)
@@ -1699,10 +1732,13 @@ void build_dyn_config(pool *p, char *_path, struct stat *stp,
 
     cp = strrchr(path, '/');
 
-    if (cp && strcmp(path, "/") != 0)
+    if (cp &&
+        strcmp(path, "/") != 0) {
       *cp = '\0';
-    else
+
+    } else {
       path = NULL;
+    }
 
     if (path) {
       if (*path && *(path + strlen(path) - 1) == '*')
@@ -2012,12 +2048,12 @@ static void _reparent_all(config_rec *newparent, xaset_t *set) {
   config_rec *c,*cnext;
 
   if (!newparent->subset)
-    newparent->subset = xaset_create(newparent->pool,NULL);
+    newparent->subset = xaset_create(newparent->pool, NULL);
 
-  for (c = (config_rec*)set->xas_list; c; c = cnext) {
+  for (c = (config_rec *) set->xas_list; c; c = cnext) {
     cnext = c->next;
-    xaset_remove(set, (xasetmember_t*)c);
-    xaset_insert(newparent->subset, (xasetmember_t*)c);
+    xaset_remove(set, (xasetmember_t *) c);
+    xaset_insert(newparent->subset, (xasetmember_t *) c);
     c->set = newparent->subset;
     c->parent = newparent;
   }
@@ -2027,20 +2063,33 @@ static void _reparent_all(config_rec *newparent, xaset_t *set) {
  * directive to.
  */
 
-static config_rec *_find_best_dir(xaset_t *set,char *path,int *matchlen)
-{
-  config_rec *c,*res = NULL,*rres;
-  int len,imatchlen,tmatchlen;
+static config_rec *_find_best_dir(xaset_t *set, char *path, size_t *matchlen) {
+  config_rec *c, *res = NULL, *rres;
+  size_t len, pathlen, imatchlen, tmatchlen;
 
   *matchlen = 0;
 
-  if (!set || !set->xas_list)
+  if (!set ||
+      !set->xas_list)
     return NULL;
 
-  for (c = (config_rec*)set->xas_list; c; c=c->next) {
+  pathlen = strlen(path);
+
+  for (c = (config_rec *) set->xas_list; c; c = c->next) {
     if (c->config_type == CONF_DIR) {
-      if (!strcmp(c->name,path))
-        continue;				/* Don't examine the current */
+      /* Note: this comparison of pointers, rather than of strings, is
+       * intentional.  DO NOT CHANGE THIS TO A strcmp()!
+       *
+       * This function is only called by reorder_dirs(), and reorder_dirs()
+       * always uses a c->name as the path parameter.  This means that
+       * doing direct pointer/address comparisons is valid.  If ever this
+       * assumption is broken, we will need to revert back to a more
+       * costly (especially when there are many <Directory> config sections)
+       * use of strcmp().
+       */
+      if (c->name == path)
+        continue;
+
       len = strlen(c->name);
       while (len > 0 &&
              (*(c->name+len-1) == '*' || *(c->name+len-1) == '/'))
@@ -2057,21 +2106,24 @@ static config_rec *_find_best_dir(xaset_t *set,char *path,int *matchlen)
        * /var/www/.1/images
        *            ^ -- /, is subdir
        */
-      if (strlen(path) > len && path[len] != '/')
+      if (pathlen > len &&
+          path[len] != '/')
           continue;
 
-      if (!strncmp(c->name,path,len) &&
-         len < strlen(path)) {
-           rres = _find_best_dir(c->subset,path,&imatchlen);
-           tmatchlen = _strmatch(path,c->name);
-           if (!rres && tmatchlen > *matchlen) {
-             res = c;
-             *matchlen = tmatchlen;
-           } else if (imatchlen > *matchlen) {
-             res = rres;
-             *matchlen = imatchlen;
-           }
-         }
+      if (len < pathlen &&
+          strncmp(c->name, path, len) == 0) {
+        rres = _find_best_dir(c->subset ,path, &imatchlen);
+        tmatchlen = _strmatch(path, c->name);
+        if (!rres &&
+            tmatchlen > *matchlen) {
+          res = c;
+          *matchlen = tmatchlen;
+
+        } else if (imatchlen > *matchlen) {
+          res = rres;
+          *matchlen = imatchlen;
+        }
+      }
     }
   }
 
@@ -2084,7 +2136,8 @@ static config_rec *_find_best_dir(xaset_t *set,char *path,int *matchlen)
 
 static void _reorder_dirs(xaset_t *set, int flags) {
   config_rec *c = NULL, *cnext = NULL, *newparent = NULL;
-  int tmp, defer = 0;
+  int defer = 0;
+  size_t tmp;
 
   if (!set || !set->xas_list)
     return;
@@ -2105,9 +2158,9 @@ static void _reorder_dirs(xaset_t *set, int flags) {
       /* If <Directory *> is used inside <Anonymous>, move all
        * the directives from '*' into the higher level.
        */
-      if (strcmp(c->name, "*") == 0 &&
-          c->parent &&
-          c->parent->config_type == CONF_ANON) {
+      if (c->parent &&
+          c->parent->config_type == CONF_ANON &&
+          strcmp(c->name, "*") == 0) {
 
         if (c->subset)
           _reparent_all(c->parent, c->subset);
@@ -2170,73 +2223,103 @@ void pr_config_dump(void (*dumpf)(const char *, ...), xaset_t *s,
   }
 }
 
-static void _mergedown(xaset_t *s,int dynamic)
-{
-  config_rec *c,*dest,*newconf;
-  int argc;
-  void **argv,**sargv;
+#ifdef PR_USE_DEVEL
+void pr_dirs_dump(void (*dumpf)(const char *, ...), xaset_t *s, char *indent) {
+  config_rec *c;
 
-  if (!s || !s->xas_list)
+  if (!s)
     return;
 
-  for (c = (config_rec*)s->xas_list; c; c=c->next) {
+  if (!indent)
+    indent = " ";
+
+  for (c = (config_rec *) s->xas_list; c; c = c->next) {
+    if (c->config_type != CONF_DIR)
+      continue;
+
+    dumpf("%s<Directory %s>", indent, c->name);
+
+    if (c->subset)
+      pr_dirs_dump(dumpf, c->subset, pstrcat(c->pool, indent, " ", NULL));
+  }
+
+  return;
+}
+#endif /* PR_USE_DEVEL */
+
+static void merge_down(xaset_t *s, int dynamic) {
+  config_rec *c, *dst, *newconf;
+  int argc;
+  void **argv, **sargv;
+
+  if (!s ||
+      !s->xas_list)
+    return;
+
+  for (c = (config_rec *) s->xas_list; c; c = c->next) {
     if ((c->flags & CF_MERGEDOWN) ||
         (c->flags & CF_MERGEDOWN_MULTI))
-      for (dest = (config_rec*)s->xas_list; dest; dest=dest->next) {
-        if (dest->config_type == CONF_ANON ||
-           dest->config_type == CONF_DIR) {
+      for (dst = (config_rec *) s->xas_list; dst; dst = dst->next) {
+        if (dst->config_type == CONF_ANON ||
+           dst->config_type == CONF_DIR) {
 
           /* If an option of the same name/type is found in the
            * next level down, it overrides, so we don't merge.
            */
           if ((c->flags & CF_MERGEDOWN) &&
-              find_config(dest->subset, c->config_type, c->name, FALSE))
+              find_config(dst->subset, c->config_type, c->name, FALSE))
             continue;
 
-          if (!dest->subset)
-            dest->subset = xaset_create(dest->pool,NULL);
+          if (!dst->subset)
+            dst->subset = xaset_create(dst->pool, NULL);
 
-          newconf = add_config_set(&dest->subset,c->name);
+          newconf = add_config_set(&dst->subset, c->name);
           newconf->config_type = c->config_type;
           newconf->flags = c->flags | (dynamic ? CF_DYNAMIC : 0);
           newconf->argc = c->argc;
-          newconf->argv = pcalloc(newconf->pool, (c->argc+1)*sizeof(void*));
-          argv = newconf->argv; sargv = c->argv;
+          newconf->argv = pcalloc(newconf->pool, (c->argc+1) * sizeof(void *));
+          argv = newconf->argv;
+          sargv = c->argv;
           argc = newconf->argc;
-          while (argc--)
+
+          while (argc--) {
             *argv++ = *sargv++;
+          }
+
           *argv++ = NULL;
         }
       }
   }
 
   /* Top level merged, recursively merge lower levels */
-  for (c = (config_rec*)s->xas_list; c; c=c->next)
-    if (c->subset && (c->config_type == CONF_ANON ||
-                     c->config_type == CONF_DIR))
-      _mergedown(c->subset,dynamic);
+  for (c = (config_rec *) s->xas_list; c; c = c->next) {
+    if (c->subset &&
+        (c->config_type == CONF_ANON ||
+         c->config_type == CONF_DIR)) {
+      merge_down(c->subset, dynamic);
+    }
+  }
 }
 
-/* iterate through <Directory> blocks inside of anonymous and
+/* Iterate through <Directory> blocks inside of anonymous and
  * resolve each one.
  */
-
-void resolve_anonymous_dirs(xaset_t *clist)
-{
+void resolve_anonymous_dirs(xaset_t *clist) {
   config_rec *c;
   char *realdir;
 
   if (!clist)
     return;
 
-  for (c = (config_rec*)clist->xas_list; c; c=c->next) {
+  for (c = (config_rec *) clist->xas_list; c; c = c->next) {
     if (c->config_type == CONF_DIR) {
       if (c->argv[1]) {
-        realdir = dir_best_path(c->pool,c->argv[1]);
-        if (realdir)
+        realdir = dir_best_path(c->pool, c->argv[1]);
+        if (realdir) {
           c->argv[1] = realdir;
-        else {
-          realdir = dir_canonical_path(c->pool,c->argv[1]);
+
+        } else {
+          realdir = dir_canonical_path(c->pool, c->argv[1]);
           if (realdir)
             c->argv[1] = realdir;
         }
@@ -2322,19 +2405,20 @@ static void copy_global_to_all(xaset_t *set) {
       copy_recur(&s->conf, s->pool, c, NULL);
 }
 
-static void fixup_globals(void) {
+static void fixup_globals(xaset_t *list) {
   server_rec *s = NULL, *smain = NULL;
   config_rec *c = NULL, *cnext = NULL;
 
-  smain = (server_rec *) server_list->xas_list;
+  smain = (server_rec *) list->xas_list;
   for (s = smain; s; s = s->next) {
     /* Loop through each top level directive looking for a CONF_GLOBAL
      * context.
      */
-    if (!s->conf || !s->conf->xas_list)
+    if (!s->conf ||
+        !s->conf->xas_list)
       continue;
 
-    for (c = (config_rec * )s->conf->xas_list; c; c = cnext) {
+    for (c = (config_rec *) s->conf->xas_list; c; c = cnext) {
       cnext = c->next;
 
       if (c->config_type == CONF_GLOBAL &&
@@ -2342,7 +2426,8 @@ static void fixup_globals(void) {
         /* Copy the contents of the block to all other servers
          * (including this one), then pull the block "out of play".
          */
-        if (c->subset && c->subset->xas_list)
+        if (c->subset &&
+            c->subset->xas_list)
           copy_global_to_all(c->subset);
 
         xaset_remove(s->conf, (xasetmember_t *) c);
@@ -2372,7 +2457,7 @@ void fixup_dirs(server_rec *s, int flags) {
   _reorder_dirs(s->conf, flags);
 
   /* Merge mergeable configuration items down. */
-  _mergedown(s->conf, FALSE);
+  merge_down(s->conf, FALSE);
 
   if (!(flags & CF_SILENT)) {
     pr_log_debug(DEBUG5, "%s", "");
@@ -2384,54 +2469,75 @@ void fixup_dirs(server_rec *s, int flags) {
 }
 
 config_rec *find_config_next(config_rec *prev, config_rec *c, int type,
-                             const char *name, int recurse)
-{
+    const char *name, int recurse) {
   config_rec *top = c;
+  unsigned int cid = 0;
 
   /* We do two searches (if recursing) so that we find the "deepest"
    * level first.
    */
 
-  if (!c && !prev)
+  if (!c &&
+      !prev)
     return NULL;
 
   if (!prev)
     prev = top;
 
+  if (name)
+    cid = pr_config_get_id(name);
+
   if (recurse) {
     do {
       config_rec *res = NULL;
 
-      for (c = top; c; c=c->next) {
-        if (c->subset && c->subset->xas_list) {
+      for (c = top; c; c = c->next) {
+        if (c->subset &&
+            c->subset->xas_list) {
           config_rec *subc = NULL;
 
-          for (subc = (config_rec *) c->subset->xas_list; subc;
-              subc = subc->next) {
-            if ((res = find_config_next(NULL, subc, type, name, recurse+1)))
+          for (subc = (config_rec *) c->subset->xas_list;
+               subc;
+               subc = subc->next) {
+            res = find_config_next(NULL, subc, type, name, recurse + 1);
+            if (res)
               return res;
           }
         }
       }
 
-      /* If deep recursion yielded no match try the current subset */
-      /* NOTE: the string comparison here is specifically case sensitive.
+      /* If deep recursion yielded no match try the current subset.
+       *
+       * NOTE: the string comparison here is specifically case sensitive.
        * The config_rec names are supplied by the modules and intentionally
        * case sensitive (they shouldn't be verbatim from the config file)
        * Do NOT change this to strcasecmp(), no matter how tempted you are
        * to do so, it will break stuff. ;)
        */
-      for (c = top; c; c=c->next) {
-        if ((type == -1 || type == c->config_type) &&
-            (!name || !strcmp(name,c->name)))
-          return c;
+      for (c = top; c; c = c->next) {
+        if (type == -1 ||
+            type == c->config_type) {
+
+          if (!name)
+            return c;
+
+          if (cid != 0 &&
+              cid == c->config_id) {
+            return c;
+          }
+
+          if (strcmp(name, c->name) == 0)
+            return c;
+        }
       }
 
       /* Restart the search at the previous level if required */
-      if (prev->parent && recurse == 1 &&
-         prev->parent->next &&
-         prev->parent->set != find_config_top) {
-        prev = top = prev->parent->next; c = top;
+      if (prev->parent &&
+          recurse == 1 &&
+          prev->parent->next &&
+          prev->parent->set != find_config_top) {
+        prev = top = prev->parent->next;
+        c = top;
         continue;
       }
 
@@ -2439,86 +2545,50 @@ config_rec *find_config_next(config_rec *prev, config_rec *c, int type,
     } while (TRUE);
 
   } else {
-    for (c = top; c; c=c->next) {
-      if ((type == -1 || type == c->config_type) &&
-         (!name || !strcmp(name, c->name)))
-        return c;
+    for (c = top; c; c = c->next) {
+      if (type == -1 ||
+          type == c->config_type) {
+
+        if (!name)
+          return c;
+
+        if (cid != 0 &&
+            cid == c->config_id) {
+          return c;
+        }
+
+        if (strcmp(name, c->name) == 0)
+          return c;
+      }
     }
   }
 
   return NULL;
 }
 
-void find_config_set_top(config_rec *c)
-{
-  if (c && c->parent)
+void find_config_set_top(config_rec *c) {
+  if (c &&
+      c->parent) {
     find_config_top = c->parent->set;
-  else
+
+  } else {
     find_config_top = NULL;
+  }
 }
 
 
-config_rec *find_config(xaset_t *set, int type, const char *name, int recurse)
-{
-  if (!set || !set->xas_list)
+config_rec *find_config(xaset_t *set, int type, const char *name, int recurse) {
+  if (!set ||
+      !set->xas_list)
     return NULL;
 
-  find_config_set_top((config_rec*)set->xas_list);
+  find_config_set_top((config_rec *) set->xas_list);
 
-  return find_config_next(NULL, (config_rec*)set->xas_list,type,name,recurse);
+  return find_config_next(NULL, (config_rec *) set->xas_list, type, name,
+    recurse);
 }
 
-/* These next two functions return the first argument in a
- * CONF_PARAM configuration entry.  If more than one or all
- * parameters are needed, the caller will need to use find_config,
- * and iterate through the argv themselves.
- * _int returns -1 if the config name is not found, _ptr returns
- * NULL.
- */
-
-long get_param_int(xaset_t *set,const char *name,int recurse)
-{
-  config_rec *c;
-
-  if (!set) {
-    _last_param_int = NULL;
-    return -1;
-  }
-
-  c = find_config(set,CONF_PARAM,name,recurse);
-
-  if (c && c->argc) {
-    _last_param_int = c;
-    return (long)c->argv[0];
-  }
-
-  _last_param_int = NULL;
-  return -1;  /* Parameters aren't allowed to contain neg. integers anyway */
-}
-
-long get_param_int_next(const char *name,int recurse)
-{
-  config_rec *c;
-
-  if (!_last_param_int || !_last_param_int->next) {
-    _last_param_int = NULL;
-    return -1;
-  }
-
-  c = find_config_next(_last_param_int,_last_param_int->next,
-                       CONF_PARAM,name,recurse);
-
-  if (c && c->argc) {
-    _last_param_int = c;
-    return (long)c->argv[0];
-  }
-
-  _last_param_int = NULL;
-  return -1;
-}
-
-void *get_param_ptr(xaset_t *set,const char *name,int recurse)
-{
+void *get_param_ptr(xaset_t *set, const char *name, int recurse) {
   config_rec *c;
 
   if (!set) {
@@ -2526,9 +2596,10 @@ void *get_param_ptr(xaset_t *set,const char *name,int recurse)
     return NULL;
   }
 
-  c = find_config(set,CONF_PARAM,name,recurse);
+  c = find_config(set, CONF_PARAM, name, recurse);
 
-  if (c && c->argc) {
+  if (c &&
+      c->argc) {
     _last_param_ptr = c;
     return c->argv[0];
   }
@@ -2540,15 +2611,17 @@ void *get_param_ptr(xaset_t *set,const char *name,int recurse)
 void *get_param_ptr_next(const char *name,int recurse) {
   config_rec *c;
 
-  if (!_last_param_ptr || !_last_param_ptr->next) {
+  if (!_last_param_ptr ||
+      !_last_param_ptr->next) {
     _last_param_ptr = NULL;
     return NULL;
   }
 
-  c = find_config_next(_last_param_ptr,_last_param_ptr->next,
-                       CONF_PARAM,name,recurse);
+  c = find_config_next(_last_param_ptr, _last_param_ptr->next, CONF_PARAM,
+    name, recurse);
 
-  if (c && c->argv) {
+  if (c &&
+      c->argv) {
     _last_param_ptr = c;
     return c->argv[0];
   }
@@ -2804,7 +2877,7 @@ int fixup_servers(xaset_t *list) {
   config_rec *c = NULL;
   server_rec *s = NULL, *next_s = NULL;
 
-  fixup_globals();
+  fixup_globals(list);
 
   s = (server_rec *) list->xas_list;
   if (s && !s->ServerName)
@@ -2830,18 +2903,21 @@ int fixup_servers(xaset_t *list) {
           const char *ipstr = pr_netaddr_get_ipstr(elts[i]);
 
 #ifdef PR_USE_IPV6
-          char ipbuf[INET6_ADDRSTRLEN];
-          if (pr_netaddr_get_family(elts[i]) == AF_INET) {
+          if (pr_netaddr_use_ipv6()) {
+            char *ipbuf = pcalloc(s->pool, INET6_ADDRSTRLEN);
+            if (pr_netaddr_get_family(elts[i]) == AF_INET) {
 
-            /* Create the bind record using the IPv4-mapped IPv6 version of
-             * this address.
-             */
-            snprintf(ipbuf, sizeof(ipbuf), "::ffff:%s", ipstr);
-            ipstr = ipbuf;
+              /* Create the bind record using the IPv4-mapped IPv6 version of
+               * this address.
+               */
+              snprintf(ipbuf, sizeof(ipbuf), "::ffff:%s", ipstr);
+              ipstr = pstrdup(s->pool, ipbuf);
+            }
           }
 #endif /* PR_USE_IPV6 */
 
-          pr_conf_add_server_config_param_str(s, "_bind", 1, ipstr);
+          if (ipstr)
+            pr_conf_add_server_config_param_str(s, "_bind", 1, ipstr);
         }
       }
  
@@ -2921,6 +2997,35 @@ void init_config(void) {
     global_config_pool = NULL;
   }
 
+  if (config_tab) {
+    /* Clear the existing config ID table.  This needs to happen when proftpd
+     * is restarting.
+     */
+    if (pr_table_empty(config_tab) < 0) {
+      pr_log_debug(DEBUG0, "error emptying config ID table: %s",
+        strerror(errno));
+    }
+
+    if (pr_table_free(config_tab) < 0) {
+      pr_log_debug(DEBUG0, "error destroying config ID table: %s",
+        strerror(errno));
+    }
+
+    config_tab = pr_table_alloc(config_tab_pool, 0);
+
+    /* Reset the ID counter as well.  Otherwise, an exceedingly long-lived
+     * proftpd, restarted many times, has the possibility of overflowing
+     * the counter data type.
+     */
+    config_id = 0;
+
+  } else {
+
+    config_tab_pool = make_sub_pool(permanent_pool);
+    pr_pool_tag(config_tab_pool, "Config ID Table Pool");
+    config_tab = pr_table_alloc(config_tab_pool, 0);
+  }
+
   if (server_list) {
     server_rec *s, *s_next;
 
@@ -3075,3 +3180,55 @@ char *get_full_cmd(cmd_rec *cmd) {
 
   return res;
 }
+
+unsigned int pr_config_get_id(const char *name) {
+  void *ptr;
+
+  if (!name) {
+    errno = EINVAL;
+    return 0;
+  }
+
+  if (!config_tab) {
+    errno = EPERM;
+    return 0;
+  }
+
+  ptr = pr_table_get(config_tab, name, 0);
+  if (!ptr) {
+    errno = ENOENT;
+    return 0;
+  }
+
+  return *((unsigned int *) ptr);
+}
+
+unsigned int pr_config_set_id(const char *name) {
+  unsigned int *ptr;
+
+  if (!name) {
+    errno = EINVAL;
+    return 0;
+  }
+
+  if (!config_tab) {
+    errno = EPERM;
+    return 0;
+  }
+
+  ptr = pr_table_pcalloc(config_tab, sizeof(unsigned int));
+  *ptr = ++config_id;
+
+  if (pr_table_add(config_tab, name, ptr, sizeof(unsigned int *)) < 0) {
+    if (errno == EEXIST) {
+      return pr_config_get_id(name);
+
+    } else {
+      pr_log_debug(DEBUG0, "error adding '%s' to config ID table: %s",
+        name, strerror(errno));
+      return 0;
+    }
+  }
+
+  return *ptr;
+}
