@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (C) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (C) 2001-2006 The ProFTPD Project
+ * Copyright (C) 2001-2007 The ProFTPD Project
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* ProFTPD virtual/modular file-system support
- * $Id: fsio.c,v 1.50 2006/03/22 22:10:34 castaglia Exp $
+ * $Id: fsio.c,v 1.64 2007/08/22 14:50:23 castaglia Exp $
  */
 
 #include "conf.h"
@@ -62,6 +62,7 @@ struct fsopendir {
   DIR *dir;
 };
 
+static const char *trace_channel = "fsio";
 static pr_fs_t *root_fs = NULL, *fs_cwd = NULL;
 static array_header *fs_map = NULL;
 
@@ -83,6 +84,9 @@ static unsigned char chk_fs_map = FALSE;
 static char vwd[PR_TUNABLE_PATH_MAX + 1] = "/";
 static char cwd[PR_TUNABLE_PATH_MAX + 1] = "/";
 
+/* Runtime enabling/disabling of UTF8 handling of paths. */
+static int use_utf8 = TRUE;
+
 /* The following static functions are simply wrappers for system functions
  */
 
@@ -298,6 +302,7 @@ typedef struct {
   char sc_path[PR_TUNABLE_PATH_MAX+1];
   struct stat sc_stat;
   int sc_errno;
+  int sc_retval;
 
 } fs_statcache_t;
 
@@ -359,7 +364,7 @@ static int cache_stat(pr_fs_t *fs, const char *path, struct stat *sbuf,
     /* Use the cached errno as well */
     errno = statcache.sc_errno;
 
-    return 0;
+    return statcache.sc_retval;
   }
 
   res = mystat(fs, pathbuf, sbuf);
@@ -369,6 +374,7 @@ static int cache_stat(pr_fs_t *fs, const char *path, struct stat *sbuf,
   sstrncpy(statcache.sc_path, pathbuf, sizeof(statcache.sc_path));
   memcpy(&statcache.sc_stat, sbuf, sizeof(struct stat));
   statcache.sc_errno = errno;
+  statcache.sc_retval = res;
 
   return res;
 }
@@ -633,6 +639,8 @@ int pr_fs_copy_file(const char *src, const char *dst) {
   pr_fsio_truncate(dst, 0);
 
   while ((res = pr_fsio_read(src_fh, buf, bufsz)) > 0) {
+    pr_signals_handle();
+
     if (pr_fsio_write(dst_fh, buf, res) != res) {
       pr_log_pri(PR_LOG_WARNING, "error copying to '%s': %s", dst,
         strerror(errno));
@@ -671,6 +679,7 @@ int pr_fs_copy_file(const char *src, const char *dst) {
 
 # elif defined(HAVE_LINUX_POSIX_ACL)
 
+#  if defined(HAVE_PERM_COPY_FD)
     /* Linux provides the handy perm_copy_fd(3) function in its libacl
      * library just for this purpose.
      */
@@ -678,6 +687,24 @@ int pr_fs_copy_file(const char *src, const char *dst) {
       pr_log_debug(DEBUG3, "error copying ACL to destination file: %s",
         strerror(errno));
 
+#  else
+    acl_t src_acl = acl_get_fd(PR_FH_FD(src_fh));
+    if (src_acl == NULL) {
+      pr_log_debug(DEBUG3, "error obtaining ACL for fd %d: %s",
+        PR_FH_FD(src_fh), strerror(errno));
+
+    } else {
+      if (acl_set_fd(PR_FH_FD(dst_fh), src_acl) < 0) {
+        pr_log_debug(DEBUG3, "error setting ACL on fd %d: %s",
+          PR_FH_FD(dst_fh), strerror(errno));
+
+      } else {
+        acl_free(src_acl);
+      }
+    }
+
+#  endif /* !HAVE_PERM_COPY_FD */
+
 # elif defined(HAVE_SOLARIS_POSIX_ACL)
     int nents;
 
@@ -734,7 +761,7 @@ pr_fs_t *pr_register_fs(pool *p, const char *name, const char *path) {
 
     /* Call pr_insert_fs() from here */
     if (!pr_insert_fs(fs, path)) {
-      pr_log_debug(DEBUG8, "FS: error inserting fs '%s' at path '%s'",
+      pr_trace_msg(trace_channel, 4, "error inserting FS '%s' at path '%s'",
         name, path);
 
       destroy_pool(fs->fs_pool);
@@ -742,7 +769,7 @@ pr_fs_t *pr_register_fs(pool *p, const char *name, const char *path) {
     }
 
   } else
-    pr_log_debug(DEBUG8, "FS: error creating fs '%s'", name);
+    pr_trace_msg(trace_channel, 6, "error creating FS '%s'", name);
 
   return fs;
 }
@@ -1295,6 +1322,11 @@ int pr_fs_dircat(char *buf, int buflen, const char *dir1, const char *dir2) {
   char *_dir1 = NULL, *_dir2 = NULL;
   size_t dir1len = 0;
 
+  if (!dir1 || !dir2) {
+    errno = EINVAL;
+    return -1;
+  }
+
   /* This is a test to see if we've got reasonable directories to concatenate.
    */
   if ((strlen(dir1) + strlen(dir2) + 1) >= PR_TUNABLE_PATH_MAX) {
@@ -1304,7 +1336,14 @@ int pr_fs_dircat(char *buf, int buflen, const char *dir1, const char *dir2) {
   }
 
   _dir1 = strdup(dir1);
+  if (!_dir1)
+    return -1;
+
   _dir2 = strdup(dir2);
+  if (!_dir2) {
+    free(_dir1);
+    return -1;
+  }
 
   dir1len = strlen(_dir1) - 1;
 
@@ -1509,9 +1548,14 @@ int pr_fs_resolve_partial(const char *path, char *buf, size_t buflen, int op) {
 
       ptr = strchr(where, '/');
 
-      if (!ptr)
-        ptr = where + strlen(where) - 1;
-      else
+      if (!ptr) {
+        size_t wherelen = strlen(where);
+
+        ptr = where;
+        if (wherelen >= 1)
+          ptr += (wherelen - 1);
+
+      } else
         *ptr = '\0';
 
       sstrncpy(namebuf, workpath, sizeof(namebuf));
@@ -1661,9 +1705,14 @@ int pr_fs_resolve_path(const char *path, char *buf, size_t buflen, int op) {
 
       ptr = strchr(where, '/');
 
-      if (!ptr)
-        ptr = where + strlen(where) - 1;
-      else
+      if (!ptr) {
+        size_t wherelen = strlen(where);
+
+        ptr = where;
+        if (wherelen >= 1)
+          ptr += (wherelen - 1);
+
+      } else
         *ptr = '\0';
 
       sstrncpy(namebuf, workpath, sizeof(namebuf));
@@ -1816,9 +1865,14 @@ void pr_fs_clean_path(const char *path, char *buf, size_t buflen) {
       }
       ptr = strchr(where, '/');
 
-      if (!ptr)
-        ptr = where + strlen(where) - 1;
-      else
+      if (!ptr) {
+        size_t wherelen = strlen(where);
+
+        ptr = where;
+        if (wherelen >= 1)
+          ptr += (wherelen - 1);
+
+      } else
         *ptr = '\0';
 
       sstrncpy(namebuf, workpath, sizeof(namebuf));
@@ -1846,6 +1900,59 @@ void pr_fs_clean_path(const char *path, char *buf, size_t buflen) {
   sstrncpy(buf, workpath, buflen);
 }
 
+int pr_fs_use_utf8(int bool) {
+  int curr_setting = use_utf8;
+  use_utf8 = bool;
+
+  return curr_setting;
+}
+
+char *pr_fs_decode_path(pool *p, const char *path) {
+#ifdef PR_USE_NLS
+  size_t outlen;
+  char *res;
+
+  if (!use_utf8) {
+    return (char *) path;
+  }
+
+  res = pr_utf8_decode(p, path, strlen(path) + 1, &outlen);
+  if (!res) {
+    pr_trace_msg("utf8", 0, "error UTF8 decoding path '%s': %s", path,
+      strerror(errno));
+    return (char *) path;
+  }
+
+  pr_trace_msg("utf8", 5, "UTF8-decoded '%s' into '%s'", path, res);
+  return res;
+#else
+  return (char *) path;
+#endif /* PR_USE_NLS */
+}
+
+char *pr_fs_encode_path(pool *p, const char *path) {
+#ifdef PR_USE_NLS
+  size_t outlen;
+  char *res;
+
+  if (!use_utf8) {
+    return (char *) path;
+  }
+
+  res = pr_utf8_encode(p, path, strlen(path) + 1, &outlen);
+  if (!res) {
+    pr_trace_msg("utf8", 0, "error UTF8 encoding path '%s': %s", path,
+      strerror(errno));
+    return (char *) path;
+  }
+
+  pr_trace_msg("utf8", 5, "UTF8-encoded '%s' into '%s'", path, res);
+  return res;
+#else
+  return (char *) path;
+#endif /* PR_USE_NLS */
+}
+
 /* This function checks the given path's prefix against the paths that
  * have been registered.  If no matching path prefix has been registered,
  * the path is considered invalid.
@@ -1938,9 +2045,14 @@ void pr_fs_virtual_path(const char *path, char *buf, size_t buflen) {
       }
       ptr = strchr(where, '/');
 
-      if (!ptr)
-        ptr = where + strlen(where) - 1;
-      else
+      if (!ptr) {
+        size_t wherelen = strlen(where);
+
+        ptr = where;
+        if (wherelen >= 1)
+          ptr += (wherelen - 1);
+
+      } else
         *ptr = '\0';
 
       sstrncpy(namebuf, workpath, sizeof(namebuf));
@@ -1984,8 +2096,9 @@ int pr_fsio_chdir_canon(const char *path, int hidesymlink) {
   while (fs && fs->fs_next && !fs->chdir)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s chdir()", fs->fs_name);
-  res = fs->chdir(fs, resbuf);
+  pr_trace_msg(trace_channel, 8, "using %s chdir() for path '%s'", fs->fs_name,
+    path);
+  res = (fs->chdir)(fs, resbuf);
 
   if (res != -1) {
     /* chdir succeeded, so we set fs_cwd for future references. */
@@ -2015,8 +2128,9 @@ int pr_fsio_chdir(const char *path, int hidesymlink) {
   while (fs && fs->fs_next && !fs->chdir)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s chdir()", fs->fs_name);
-  res = fs->chdir(fs, resbuf);
+  pr_trace_msg(trace_channel, 8, "using %s chdir() for path '%s'", fs->fs_name,
+    path);
+  res = (fs->chdir)(fs, resbuf);
 
   if (res != -1) {
     /* chdir succeeded, so we set fs_cwd for future references. */
@@ -2059,8 +2173,9 @@ void *pr_fsio_opendir(const char *path) {
   while (fs && fs->fs_next && !fs->opendir)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s opendir()", fs->fs_name);
-  res = fs->opendir(fs, path);
+  pr_trace_msg(trace_channel, 8, "using %s opendir() for path '%s'",
+    fs->fs_name, path);
+  res = (fs->opendir)(fs, path);
 
   if (!res)
     return NULL;
@@ -2077,7 +2192,7 @@ void *pr_fsio_opendir(const char *path) {
   if (!fsod) {
 
     if (fs->closedir) {
-      fs->closedir(fs, res);
+      (fs->closedir)(fs, res);
       errno = ENOMEM;
       return NULL;
 
@@ -2165,8 +2280,8 @@ int pr_fsio_closedir(void *dir) {
   while (fs && fs->fs_next && !fs->closedir)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s closedir()", fs->fs_name);
-  res = fs->closedir(fs, dir);
+  pr_trace_msg(trace_channel, 8, "using %s closedir()", fs->fs_name);
+  res = (fs->closedir)(fs, dir);
 
   return res;
 }
@@ -2184,8 +2299,8 @@ struct dirent *pr_fsio_readdir(void *dir) {
   while (fs && fs->fs_next && !fs->readdir)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s readdir()", fs->fs_name);
-  res = fs->readdir(fs, dir);
+  pr_trace_msg(trace_channel, 8, "using %s readdir()", fs->fs_name);
+  res = (fs->readdir)(fs, dir);
 
   return res;
 }
@@ -2200,8 +2315,9 @@ int pr_fsio_mkdir(const char *path, mode_t mode) {
   while (fs && fs->fs_next && !fs->mkdir)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s mkdir()", fs->fs_name);
-  res = fs->mkdir(fs, path, mode);
+  pr_trace_msg(trace_channel, 8, "using %s mkdir() for path '%s'", fs->fs_name,
+    path);
+  res = (fs->mkdir)(fs, path, mode);
 
   return res;
 }
@@ -2216,8 +2332,9 @@ int pr_fsio_rmdir(const char *path) {
   while (fs && fs->fs_next && !fs->rmdir)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s rmdir()", fs->fs_name);
-  res = fs->rmdir(fs, path);
+  pr_trace_msg(trace_channel, 8, "using %s rmdir() for path '%s'", fs->fs_name,
+    path);
+  res = (fs->rmdir)(fs, path);
 
   return res;
 }
@@ -2231,7 +2348,8 @@ int pr_fsio_stat_canon(const char *path, struct stat *sbuf) {
   while (fs && fs->fs_next && !fs->stat)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s stat()", fs ? fs->fs_name : "system");
+  pr_trace_msg(trace_channel, 8, "using %s stat() for path '%s'",
+    fs ? fs->fs_name : "system", path);
   return fs_cache_stat(fs ? fs : root_fs, path, sbuf);
 }
 
@@ -2244,7 +2362,8 @@ int pr_fsio_stat(const char *path, struct stat *sbuf) {
   while (fs && fs->fs_next && !fs->stat)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s stat()", fs->fs_name);
+  pr_trace_msg(trace_channel, 8, "using %s stat() for path '%s'", fs->fs_name,
+    path);
   return fs_cache_stat(fs ? fs : root_fs, path, sbuf);
 }
 
@@ -2264,8 +2383,9 @@ int pr_fsio_fstat(pr_fh_t *fh, struct stat *sbuf) {
   while (fs && fs->fs_next && !fs->fstat)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s fstat()", fs->fs_name);
-  res = fs->fstat(fh, fh->fh_fd, sbuf);
+  pr_trace_msg(trace_channel, 8, "using %s fstat() for path '%s'", fs->fs_name,
+    fh->fh_path);
+  res = (fs->fstat)(fh, fh->fh_fd, sbuf);
 
   return res;
 }
@@ -2279,7 +2399,8 @@ int pr_fsio_lstat_canon(const char *path, struct stat *sbuf) {
   while (fs && fs->fs_next && !fs->lstat)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s lstat()", fs ? fs->fs_name : "system");
+  pr_trace_msg(trace_channel, 8, "using %s lstat() for path '%s'",
+    fs ? fs->fs_name : "system", path);
   return fs_cache_lstat(fs ? fs : root_fs, path, sbuf);
 }
 
@@ -2292,7 +2413,8 @@ int pr_fsio_lstat(const char *path, struct stat *sbuf) {
   while (fs && fs->fs_next && !fs->lstat)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s lstat()", fs->fs_name);
+  pr_trace_msg(trace_channel, 8, "using %s lstat() for path '%s'", fs->fs_name,
+    path);
   return fs_cache_lstat(fs ? fs : root_fs, path, sbuf);
 }
 
@@ -2306,8 +2428,9 @@ int pr_fsio_readlink_canon(const char *path, char *buf, size_t buflen) {
   while (fs && fs->fs_next && !fs->readlink)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s readlink()", fs->fs_name);
-  res = fs->readlink(fs, path, buf, buflen);
+  pr_trace_msg(trace_channel, 8, "using %s readlink() for path '%s'",
+    fs->fs_name, path);
+  res = (fs->readlink)(fs, path, buf, buflen);
 
   return res;
 }
@@ -2322,8 +2445,9 @@ int pr_fsio_readlink(const char *path, char *buf, size_t buflen) {
   while (fs && fs->fs_next && !fs->readlink)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s readlink()", fs->fs_name);
-  res = fs->readlink(fs, path, buf, buflen);
+  pr_trace_msg(trace_channel, 8, "using %s readlink() for path '%s'",
+    fs->fs_name, path);
+  res = (fs->readlink)(fs, path, buf, buflen);
 
   return res;
 }
@@ -2366,8 +2490,9 @@ int pr_fsio_rename_canon(const char *rfrom, const char *rto) {
   while (fs && fs->fs_next && !fs->rename)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s rename()", fs->fs_name);
-  res = fs->rename(fs, rfrom, rto);
+  pr_trace_msg(trace_channel, 8, "using %s rename() for paths '%s', '%s'",
+    fs->fs_name, rfrom, rto);
+  res = (fs->rename)(fs, rfrom, rto);
 
   return res;
 }
@@ -2387,8 +2512,9 @@ int pr_fsio_rename(const char *rnfm, const char *rnto) {
   while (fs && fs->fs_next && !fs->rename)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s rename()", fs->fs_name);
-  res = fs->rename(fs, rnfm, rnto);
+  pr_trace_msg(trace_channel, 8, "using %s rename() for paths '%s', '%s'",
+    fs->fs_name, rnfm, rnto);
+  res = (fs->rename)(fs, rnfm, rnto);
 
   return res;
 }
@@ -2403,8 +2529,9 @@ int pr_fsio_unlink_canon(const char *name) {
   while (fs && fs->fs_next && !fs->unlink)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s unlink()", fs->fs_name);
-  res = fs->unlink(fs, name);
+  pr_trace_msg(trace_channel, 8, "using %s unlink() for path '%s'",
+    fs->fs_name, name);
+  res = (fs->unlink)(fs, name);
 
   return res;
 }
@@ -2419,8 +2546,9 @@ int pr_fsio_unlink(const char *name) {
   while (fs && fs->fs_next && !fs->unlink)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s unlink()", fs->fs_name);
-  res = fs->unlink(fs, name);
+  pr_trace_msg(trace_channel, 8, "using %s unlink() for path '%s'",
+    fs->fs_name, name);
+  res = (fs->unlink)(fs, name);
 
   return res;
 }
@@ -2449,8 +2577,9 @@ pr_fh_t *pr_fsio_open_canon(const char *name, int flags) {
   while (fs && fs->fs_next && !fs->open)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s open()", fs->fs_name);
-  fh->fh_fd = fs->open(fh, deref, flags);
+  pr_trace_msg(trace_channel, 8, "using %s open() for path '%s'", fs->fs_name,
+    name);
+  fh->fh_fd = (fs->open)(fh, deref, flags);
 
   if (fh->fh_fd == -1) {
     destroy_pool(fh->fh_pool);
@@ -2489,8 +2618,9 @@ pr_fh_t *pr_fsio_open(const char *name, int flags) {
   while (fs && fs->fs_next && !fs->open)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s open()", fs->fs_name);
-  fh->fh_fd = fs->open(fh, name, flags);
+  pr_trace_msg(trace_channel, 8, "using %s open() for path '%s'", fs->fs_name,
+    name);
+  fh->fh_fd = (fs->open)(fh, name, flags);
 
   if (fh->fh_fd == -1) {
     destroy_pool(fh->fh_pool);
@@ -2523,8 +2653,9 @@ pr_fh_t *pr_fsio_creat_canon(const char *name, mode_t mode) {
   while (fs && fs->fs_next && !fs->creat)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s creat()", fs->fs_name);
-  fh->fh_fd = fs->creat(fh, deref, mode);
+  pr_trace_msg(trace_channel, 8, "using %s creat() for path '%s'", fs->fs_name,
+    name);
+  fh->fh_fd = (fs->creat)(fh, deref, mode);
 
   if (fh->fh_fd == -1) {
     destroy_pool(fh->fh_pool);
@@ -2556,8 +2687,9 @@ pr_fh_t *pr_fsio_creat(const char *name, mode_t mode) {
   while (fs && fs->fs_next && !fs->creat)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s creat()", fs->fs_name);
-  fh->fh_fd = fs->creat(fh, name, mode);
+  pr_trace_msg(trace_channel, 8, "using %s creat() for path '%s'", fs->fs_name,
+    name);
+  fh->fh_fd = (fs->creat)(fh, name, mode);
 
   if (fh->fh_fd == -1) {
     destroy_pool(fh->fh_pool);
@@ -2583,8 +2715,9 @@ int pr_fsio_close(pr_fh_t *fh) {
   while (fs && fs->fs_next && !fs->close)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s close()", fs->fs_name);
-  res = fs->close(fh, fh->fh_fd);
+  pr_trace_msg(trace_channel, 8, "using %s close() for path '%s'", fs->fs_name,
+    fh->fh_path);
+  res = (fs->close)(fh, fh->fh_fd);
 
   destroy_pool(fh->fh_pool);
   return res;
@@ -2606,8 +2739,9 @@ int pr_fsio_read(pr_fh_t *fh, char *buf, size_t size) {
   while (fs && fs->fs_next && !fs->read)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s read()", fs->fs_name);
-  res = fs->read(fh, fh->fh_fd, buf, size);
+  pr_trace_msg(trace_channel, 8, "using %s read() for path '%s'", fs->fs_name,
+    fh->fh_path);
+  res = (fs->read)(fh, fh->fh_fd, buf, size);
 
   return res;
 }
@@ -2628,8 +2762,9 @@ int pr_fsio_write(pr_fh_t *fh, const char *buf, size_t size) {
   while (fs && fs->fs_next && !fs->write)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s write()", fs->fs_name);
-  res = fs->write(fh, fh->fh_fd, buf, size);
+  pr_trace_msg(trace_channel, 8, "using %s write() for path '%s'", fs->fs_name,
+    fh->fh_path);
+  res = (fs->write)(fh, fh->fh_fd, buf, size);
 
   return res;
 }
@@ -2650,8 +2785,9 @@ off_t pr_fsio_lseek(pr_fh_t *fh, off_t offset, int whence) {
   while (fs && fs->fs_next && !fs->lseek)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s lseek()", fs->fs_name);
-  res = fs->lseek(fh, fh->fh_fd, offset, whence);
+  pr_trace_msg(trace_channel, 8, "using %s lseek() for path '%s'", fs->fs_name,
+    fh->fh_path);
+  res = (fs->lseek)(fh, fh->fh_fd, offset, whence);
 
   return res;
 }
@@ -2671,8 +2807,9 @@ int pr_fsio_link_canon(const char *lfrom, const char *lto) {
   while (fs && fs->fs_next && !fs->link)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s link()", fs->fs_name);
-  res = fs->link(fs, lfrom, lto);
+  pr_trace_msg(trace_channel, 8, "using %s link() for paths '%s', '%s'",
+    fs->fs_name, lfrom, lto);
+  res = (fs->link)(fs, lfrom, lto);
 
   return res;
 }
@@ -2692,8 +2829,9 @@ int pr_fsio_link(const char *lfrom, const char *lto) {
   while (fs && fs->fs_next && !fs->link)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s link()", fs->fs_name);
-  res = fs->link(fs, lfrom, lto);
+  pr_trace_msg(trace_channel, 8, "using %s link() for paths '%s', '%s'",
+    fs->fs_name, lfrom, lto);
+  res = (fs->link)(fs, lfrom, lto);
 
   return res;
 }
@@ -2708,8 +2846,9 @@ int pr_fsio_symlink_canon(const char *lfrom, const char *lto) {
   while (fs && fs->fs_next && !fs->symlink)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s symlink()", fs->fs_name);
-  res = fs->symlink(fs, lfrom, lto);
+  pr_trace_msg(trace_channel, 8, "using %s symlink() for path '%s'",
+    fs->fs_name, lto);
+  res = (fs->symlink)(fs, lfrom, lto);
 
   return res;
 }
@@ -2724,8 +2863,9 @@ int pr_fsio_symlink(const char *lfrom, const char *lto) {
   while (fs && fs->fs_next && !fs->symlink)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s symlink()", fs->fs_name);
-  res = fs->symlink(fs, lfrom, lto);
+  pr_trace_msg(trace_channel, 8, "using %s symlink() for path '%s'",
+    fs->fs_name, lto);
+  res = (fs->symlink)(fs, lfrom, lto);
 
   return res;
 }
@@ -2746,8 +2886,9 @@ int pr_fsio_ftruncate(pr_fh_t *fh, off_t len) {
   while (fs && fs->fs_next && !fs->ftruncate)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s ftruncate()", fs->fs_name);
-  res = fs->ftruncate(fh, fh->fh_fd, len);
+  pr_trace_msg(trace_channel, 8, "using %s ftruncate() for path '%s'",
+    fs->fs_name, fh->fh_path);
+  res = (fs->ftruncate)(fh, fh->fh_fd, len);
 
   return res;
 }
@@ -2762,8 +2903,9 @@ int pr_fsio_truncate_canon(const char *path, off_t len) {
   while (fs && fs->fs_next && !fs->truncate)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s truncate()", fs->fs_name);
-  res = fs->truncate(fs, path, len);
+  pr_trace_msg(trace_channel, 8, "using %s truncate() for path '%s'",
+    fs->fs_name, path);
+  res = (fs->truncate)(fs, path, len);
 
   return res;
 }
@@ -2778,8 +2920,9 @@ int pr_fsio_truncate(const char *path, off_t len) {
   while (fs && fs->fs_next && !fs->truncate)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s truncate()", fs->fs_name);
-  res = fs->truncate(fs, path, len);
+  pr_trace_msg(trace_channel, 8, "using %s truncate() for path '%s'",
+    fs->fs_name, path);
+  res = (fs->truncate)(fs, path, len);
   
   return res;
 }
@@ -2795,8 +2938,12 @@ int pr_fsio_chmod_canon(const char *name, mode_t mode) {
   while (fs && fs->fs_next && !fs->chmod)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s chmod()", fs->fs_name);
-  res = fs->chmod(fs, deref, mode);
+  pr_trace_msg(trace_channel, 8, "using %s chmod() for path '%s'",
+    fs->fs_name, name);
+  res = (fs->chmod)(fs, deref, mode);
+
+  if (res == 0)
+    pr_fs_clear_cache();
 
   return res;
 }
@@ -2811,9 +2958,13 @@ int pr_fsio_chmod(const char *name, mode_t mode) {
   while (fs && fs->fs_next && !fs->chmod)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s chmod()", fs->fs_name);
-  res = fs->chmod(fs, name, mode);
-  
+  pr_trace_msg(trace_channel, 8, "using %s chmod() for path '%s'",
+    fs->fs_name, name);
+  res = (fs->chmod)(fs, name, mode);
+
+  if (res == 0)
+    pr_fs_clear_cache();
+
   return res;
 }
 
@@ -2827,9 +2978,13 @@ int pr_fsio_chown_canon(const char *name, uid_t uid, gid_t gid) {
   while (fs && fs->fs_next && !fs->chown)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s chown()", fs->fs_name);
-  res = fs->chown(fs, name, uid, gid);
-  
+  pr_trace_msg(trace_channel, 8, "using %s chown() for path '%s'",
+    fs->fs_name, name);
+  res = (fs->chown)(fs, name, uid, gid);
+
+  if (res == 0)
+    pr_fs_clear_cache();
+
   return res;
 }
 
@@ -2843,9 +2998,13 @@ int pr_fsio_chown(const char *name, uid_t uid, gid_t gid) {
   while (fs && fs->fs_next && !fs->chown)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s chown()", fs->fs_name);
-  res = fs->chown(fs, name, uid, gid);
-  
+  pr_trace_msg(trace_channel, 8, "using %s chown() for path '%s'",
+    fs->fs_name, name);
+  res = (fs->chown)(fs, name, uid, gid);
+
+  if (res == 0)
+    pr_fs_clear_cache();
+
   return res;
 }
 
@@ -2866,8 +3025,9 @@ int pr_fsio_access(const char *path, int mode, uid_t uid, gid_t gid,
   while (fs && fs->fs_next && !fs->access)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s access()", fs->fs_name);
-  return fs->access(fs, path, mode, uid, gid, suppl_gids);
+  pr_trace_msg(trace_channel, 8, "using %s access() for path '%s'",
+    fs->fs_name, path);
+  return (fs->access)(fs, path, mode, uid, gid, suppl_gids);
 }
 
 int pr_fsio_faccess(pr_fh_t *fh, int mode, uid_t uid, gid_t gid,
@@ -2886,8 +3046,9 @@ int pr_fsio_faccess(pr_fh_t *fh, int mode, uid_t uid, gid_t gid,
   while (fs && fs->fs_next && !fs->faccess)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s faccess()", fs->fs_name);
-  return fh->fh_fs->faccess(fh, mode, uid, gid, suppl_gids);
+  pr_trace_msg(trace_channel, 8, "using %s faccess() for path '%s'",
+    fs->fs_name, fh->fh_path);
+  return (fh->fh_fs->faccess)(fh, mode, uid, gid, suppl_gids);
 }
 
 /* If the wrapped chroot() function suceeds (eg returns 0), then all
@@ -2904,8 +3065,9 @@ int pr_fsio_chroot(const char *path) {
   while (fs && fs->fs_next && !fs->chroot)
     fs = fs->fs_next;
 
-  pr_log_debug(DEBUG9, "FS: using %s chroot()", fs->fs_name);
-  res = fs->chroot(fs, path);
+  pr_trace_msg(trace_channel, 8, "using %s chroot() for path '%s'",
+    fs->fs_name, path);
+  res = (fs->chroot)(fs, path);
   if (res == 0) {
     unsigned int iter_start = 0;
 
@@ -2950,6 +3112,8 @@ int pr_fsio_chroot(const char *path) {
         /* Need to do this for any stacked FSs as well. */
         next = tmpfs->fs_next;
         while (next) {
+          pr_signals_handle();
+
           memmove(next->fs_path, next->fs_path + strlen(path),
             strlen(next->fs_path) - strlen(path) + 1);
 
@@ -2996,6 +3160,8 @@ char *pr_fsio_gets(char *buf, size_t size, pr_fh_t *fh) {
   bp = buf;
 
   while (size) {
+    pr_signals_handle();
+
     if (!pbuf->current ||
         pbuf->remaining == pbuf->buflen) { /* empty buffer */
 
@@ -3017,13 +3183,20 @@ char *pr_fsio_gets(char *buf, size_t size, pr_fh_t *fh) {
     } else
       toread = pbuf->buflen - pbuf->remaining;
 
-    while (size && toread > 0 && *pbuf->current != '\n' && toread--) {
+    while (size &&
+           toread > 0 &&
+           *pbuf->current != '\n' &&
+           toread--) {
+      pr_signals_handle();
+
       *bp++ = *pbuf->current++;
       size--;
       pbuf->remaining++;
     }
 
-    if (size && toread && *pbuf->current == '\n') {
+    if (size &&
+        toread &&
+        *pbuf->current == '\n') {
       size--;
       toread--;
       *bp++ = *pbuf->current++;
@@ -3051,8 +3224,6 @@ char *pr_fsio_getline(char *buf, int buflen, pr_fh_t *fh,
   char *start = buf;
 
   while (pr_fsio_gets(buf, buflen, fh)) {
-
-    /* while() loops should always handle signals. */
     pr_signals_handle();
 
     inlen = strlen(buf);
@@ -3269,3 +3440,127 @@ int init_fs(void) {
 
   return 0;
 }
+
+#ifdef PR_USE_DEVEL
+
+static const char *get_fs_hooks_str(pool *p, pr_fs_t *fs) {
+  char *hooks = "";
+
+  if (fs->stat)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "stat(2)", NULL);
+
+  if (fs->lstat)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "lstat(2)", NULL);
+
+  if (fs->fstat)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "fstat(2)", NULL);
+
+  if (fs->rename)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "rename(2)", NULL);
+
+  if (fs->link)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "link(2)", NULL);
+
+  if (fs->unlink)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "unlink(2)", NULL);
+
+  if (fs->open)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "open(2)", NULL);
+
+  if (fs->creat)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "creat(2)", NULL);
+
+  if (fs->close)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "close(2)", NULL);
+
+  if (fs->read)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "read(2)", NULL);
+
+  if (fs->lseek)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "lseek(2)", NULL);
+
+  if (fs->readlink)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "readlink(2)", NULL);
+
+  if (fs->symlink)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "symlink(2)", NULL);
+
+  if (fs->ftruncate)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "ftruncate(2)", NULL);
+
+  if (fs->truncate)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "truncate(2)", NULL);
+
+  if (fs->chmod)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "chmod(2)", NULL);
+
+  if (fs->chown)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "chown(2)", NULL);
+
+  if (fs->access)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "access(2)", NULL);
+
+  if (fs->faccess)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "faccess(2)", NULL);
+
+  if (fs->chdir)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "chdir(2)", NULL);
+
+  if (fs->chroot)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "chroot(2)", NULL);
+
+  if (fs->opendir)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "opendir(3)", NULL);
+
+  if (fs->closedir)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "closedir(3)", NULL);
+
+  if (fs->readdir)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "readdir(3)", NULL);
+
+  if (fs->mkdir)
+    hooks = pstrcat(p, hooks, *hooks ? ", " : "", "mkdir(2)", NULL);
+
+  if (!*hooks) {
+    return pstrdup(p, "(none)");
+  }
+
+  return hooks;
+}
+
+static void get_fs_info(pool *p, int depth, pr_fs_t *fs,
+    void (*dumpf)(const char *, ...)) {
+
+  dumpf("FS#%u: '%s', mounted at '%s', implementing the following hooks:",
+    depth, fs->fs_name, fs->fs_path);
+  dumpf("FS#%u:    %s", depth, get_fs_hooks_str(p, fs));
+}
+
+void pr_fs_dump(void (*dumpf)(const char *, ...)) {
+  pool *p;
+
+  dumpf("FS#0: 'system' mounted at '/', implementing the following hooks:");
+  dumpf("FS#0:    (all)");
+
+  if (!fs_map ||
+      fs_map->nelts == 0)
+    return;
+
+  p = make_sub_pool(permanent_pool);
+
+  if (fs_map->nelts > 0) {
+    pr_fs_t **fs_objs = (pr_fs_t **) fs_map->elts;
+    register int i;
+
+    for (i = 0; i < fs_map->nelts; i++) {
+      pr_fs_t *fsi = fs_objs[i];
+
+      for (; fsi->fs_next; fsi = fsi->fs_next) {
+        get_fs_info(p, i+1, fsi, dumpf);
+      }
+    }
+  }
+
+  destroy_pool(p);
+}
+#endif /* PR_USE_DEVEL */
