@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2005 The ProFTPD Project team
+ * Copyright (c) 2001-2007 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,7 @@
  */
 
 /* ProFTPD logging support.
- * $Id: log.c,v 1.76 2005/11/14 16:33:06 castaglia Exp $
+ * $Id: log.c,v 1.82 2007/09/11 00:49:44 castaglia Exp $
  */
 
 #include "conf.h"
@@ -63,7 +63,7 @@ int log_wtmp(char *line, const char *name, const char *host,
     (defined(__NetBSD__) && defined(HAVE_UTMPX_H))) && \
     !(defined(LINUX) || defined(__hpux) || defined (_AIX))
   /* This "auxilliary" utmp doesn't exist under linux. */
-#ifdef __sparcv9
+#if defined(__sparcv9) && !defined(__NetBSD__)
   struct futmpx utx;
   time_t t;
 #else
@@ -117,7 +117,7 @@ int log_wtmp(char *line, const char *name, const char *host,
     utx.ut_exit.e_exit = 0;
 #endif /* HAVE_UT_UT_EXIT */
     if (write(fdx, (char *)&utx, sizeof(utx)) != sizeof(utx))
-      ftruncate(fdx, buf.st_size);
+      (void) ftruncate(fdx, buf.st_size);
 
   } else {
     pr_log_debug(DEBUG0, "%s fstat(): %s", WTMPX_FILE, strerror(errno));
@@ -357,7 +357,9 @@ int pr_log_writefile(int logfd, const char *ident, const char *fmt, ...) {
     return -1;
   }
 
-  t = localtime(&timestamp);
+  t = pr_localtime(NULL, &timestamp);
+  if (!t) 
+    return -1;
 
   /* Prepend the timestamp */
   strftime(buf, sizeof(buf), "%b %d %H:%M:%S ", t);
@@ -376,8 +378,14 @@ int pr_log_writefile(int logfd, const char *ident, const char *fmt, ...) {
   buf[sizeof(buf)-1] = '\0';
   buf[strlen(buf)] = '\n';
 
-  if (write(logfd, buf, strlen(buf)) < 0)
+  while (write(logfd, buf, strlen(buf)) < 0) {
+    if (errno == EINTR) {
+      pr_signals_handle();
+      continue;
+    }
+
     return -1;
+  }
 
   return 0;
 }
@@ -471,7 +479,10 @@ static void log_write(int priority, int f, char *s) {
     time_t tt = time(NULL);
     struct tm *t;
 
-    t = localtime(&tt);
+    t = pr_localtime(NULL, &tt);
+    if (!t)
+      return;
+
     strftime(buf, sizeof(buf), "%b %d %H:%M:%S ", t);
     buf[sizeof(buf) - 1] = '\0';
 
@@ -486,7 +497,14 @@ static void log_write(int priority, int f, char *s) {
     }
 
     buf[sizeof(buf) - 1] = '\0';
-    write(systemlog_fd, buf, strlen(buf));
+    while (write(systemlog_fd, buf, strlen(buf)) < 0) {
+      if (errno == EINTR) {
+        pr_signals_handle();
+        continue;
+      }
+
+      return;
+    }
     return;
   }
 
