@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2007 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,7 +23,7 @@
  */
 
 /* NetIO routines
- * $Id: netio.c,v 1.21 2004/10/09 20:46:22 castaglia Exp $
+ * $Id: netio.c,v 1.28 2007/08/22 14:50:23 castaglia Exp $
  */
 
 #include "conf.h"
@@ -45,6 +45,8 @@
 #define WILL	251
 #endif
 
+static const char *trace_channel = "netio";
+
 static pr_netio_t *core_ctrl_netio = NULL, *ctrl_netio = NULL;
 static pr_netio_t *core_data_netio = NULL, *data_netio = NULL;
 static pr_netio_t *core_othr_netio = NULL, *othr_netio = NULL;
@@ -128,11 +130,16 @@ static int core_netio_poll_cb(pr_netio_stream_t *nstrm) {
   FD_ZERO(&rfds);
   FD_ZERO(&wfds);
 
-  if (nstrm->strm_mode == PR_NETIO_IO_RD)
-    FD_SET(nstrm->strm_fd, &rfds);
+  if (nstrm->strm_mode == PR_NETIO_IO_RD) {
+    if (nstrm->strm_fd >= 0) {
+      FD_SET(nstrm->strm_fd, &rfds);
+    }
 
-  else
-    FD_SET(nstrm->strm_fd, &wfds);
+  } else {
+    if (nstrm->strm_fd >= 0) {
+      FD_SET(nstrm->strm_fd, &wfds);
+    }
+  }
 
   tval.tv_sec = ((nstrm->strm_flags & PR_NETIO_SESS_INTR) ?
     nstrm->strm_interval: 60);
@@ -231,58 +238,71 @@ int pr_netio_close(pr_netio_stream_t *nstrm) {
 
 static int netio_lingering_close(pr_netio_stream_t *nstrm, long linger,
     int flags) {
-
   int res;
-  struct timeval tv;
-  fd_set rs;
-  time_t when = time(NULL) + linger;
 
   if (!nstrm) {
     errno = EINVAL;
     return -1;
   }
 
+  if (nstrm->strm_fd < 0)
+    /* Already closed. */
+    return 0;
+
   if (!(flags & NETIO_LINGERING_CLOSE_FL_NO_SHUTDOWN))
     pr_netio_shutdown(nstrm, 1);
 
-  tv.tv_sec = linger;
-  tv.tv_usec = 0L;
+  if (nstrm->strm_fd >= 0) {
+    struct timeval tv;
+    fd_set rs;
+    time_t when = time(NULL) + linger;
 
-  /* Handle timers during reading, once selected for read this
-   * should mean all buffers have been flushed and the receiving end
-   * has closed.
-   */
-  while (TRUE) {
-    run_schedule();
+    tv.tv_sec = linger;
+    tv.tv_usec = 0L;
 
-    FD_ZERO(&rs);
-    FD_SET(nstrm->strm_fd, &rs);
+    /* Handle timers during reading, once selected for read this
+     * should mean all buffers have been flushed and the receiving end
+     * has closed.
+     */
+    while (TRUE) {
+      run_schedule();
 
-    res = select(nstrm->strm_fd+1, &rs, NULL, NULL, &tv);
-    if (res == -1) {
-      if (errno == EINTR) {
-        time_t now = time(NULL);
-        pr_signals_handle();
+      FD_ZERO(&rs);
+      FD_SET(nstrm->strm_fd, &rs);
 
-        /* Still here? If the requested lingering interval hasn't passed,
-         * continue lingering.  Reset the timeval struct's fields to
-         * linger for the interval remaining in the given period of time.
-         */
-        if (now < when) {
-          tv.tv_sec = when - now;
-          tv.tv_usec = 0L;
-          continue;
+      pr_trace_msg(trace_channel, 8,
+        "lingering %lu secs before closing fd %d", (unsigned long) tv.tv_sec,
+        nstrm->strm_fd);
+
+      res = select(nstrm->strm_fd+1, &rs, NULL, NULL, &tv);
+      if (res == -1) {
+        if (errno == EINTR) {
+          time_t now = time(NULL);
+          pr_signals_handle();
+
+          /* Still here? If the requested lingering interval hasn't passed,
+           * continue lingering.  Reset the timeval struct's fields to
+           * linger for the interval remaining in the given period of time.
+           */
+          if (now < when) {
+            tv.tv_sec = when - now;
+            tv.tv_usec = 0L;
+            continue;
+          }
+
+        } else {
+          nstrm->strm_errno = errno;
+          return -1;
         }
-
-      } else {
-        nstrm->strm_errno = errno;
-        return -1;
       }
-    }
 
-    break;
+      break;
+    }
   }
 
+  pr_trace_msg(trace_channel, 8, "done lingering, closing fd %d",
+    nstrm->strm_fd);
+
   if (nstrm->strm_type == PR_NETIO_STRM_CTRL)
     return ctrl_netio ? ctrl_netio->close(nstrm) :
       core_ctrl_netio->close(nstrm);
@@ -301,48 +321,51 @@ static int netio_lingering_close(pr_netio_stream_t *nstrm, long linger,
 
 int pr_netio_lingering_abort(pr_netio_stream_t *nstrm, long linger) {
   int res;
-  struct timeval tv;
-  fd_set rs;
 
   if (!nstrm) {
     errno = EINVAL;
     return -1;
   }
 
+  /* Send an appropriate response code down the stream asychronously. */
+  pr_response_send_async(R_426, _("Transfer aborted. Data connection closed."));
+
   pr_netio_shutdown(nstrm, 1);
 
-  /* Wait for just a little while for the shutdown to take effect. */
-  tv.tv_sec = 0L;
-  tv.tv_usec = 300000L;
+  if (nstrm->strm_fd >= 0) {
+    fd_set rs;
+    struct timeval tv;
 
-  while (TRUE) {
-    run_schedule();
+    /* Wait for just a little while for the shutdown to take effect. */
+    tv.tv_sec = 0L;
+    tv.tv_usec = 300000L;
 
-    FD_ZERO(&rs);
-    FD_SET(nstrm->strm_fd, &rs);
+    while (TRUE) {
+      run_schedule();
 
-    res = select(nstrm->strm_fd+1, &rs, NULL, NULL, &tv);
-    if (res == -1) {
-      if (errno == EINTR) {
-        pr_signals_handle();
+      FD_ZERO(&rs);
+      FD_SET(nstrm->strm_fd, &rs);
 
-        /* Linger some more. */
-        tv.tv_sec = 0L;
-        tv.tv_usec = 300000L;
-        continue;
+      res = select(nstrm->strm_fd+1, &rs, NULL, NULL, &tv);
+      if (res == -1) {
+        if (errno == EINTR) {
+          pr_signals_handle();
 
-      } else {
-        nstrm->strm_errno = errno;
-        return -1;
+          /* Linger some more. */
+          tv.tv_sec = 0L;
+          tv.tv_usec = 300000L;
+          continue;
+
+        } else {
+          nstrm->strm_errno = errno;
+          return -1;
+        }
       }
+
+      break;
     }
-
-    break;
   }
 
-  /* Send an appropriate response code down the stream asychronously. */
-  pr_response_send_async(R_426, "Transfer aborted. Data connection closed.");
-
   /* Now continue with a normal lingering close. */
   return netio_lingering_close(nstrm, linger,
     NETIO_LINGERING_CLOSE_FL_NO_SHUTDOWN);  
@@ -367,25 +390,26 @@ pr_netio_stream_t *pr_netio_open(pool *parent_pool, int strm_type, int fd,
   if (strm_type == PR_NETIO_STRM_CTRL) {
     nstrm->strm_type = PR_NETIO_STRM_CTRL;
     nstrm->strm_mode = mode;
-    return ctrl_netio ? ctrl_netio->open(nstrm, fd, mode) :
-      core_ctrl_netio->open(nstrm, fd, mode);
+    return ctrl_netio ? (ctrl_netio->open)(nstrm, fd, mode) :
+      (core_ctrl_netio->open)(nstrm, fd, mode);
   }
 
   if (strm_type == PR_NETIO_STRM_DATA) {
     nstrm->strm_type = PR_NETIO_STRM_DATA;
     nstrm->strm_mode = mode;
-    return data_netio ? data_netio->open(nstrm, fd, mode) :
-      core_data_netio->open(nstrm, fd, mode);
+    return data_netio ? (data_netio->open)(nstrm, fd, mode) :
+      (core_data_netio->open)(nstrm, fd, mode);
   }
 
   if (strm_type == PR_NETIO_STRM_OTHR) {
     nstrm->strm_type = PR_NETIO_STRM_OTHR;
     nstrm->strm_mode = mode;
-    return othr_netio ? othr_netio->open(nstrm, fd, mode) :
-      core_othr_netio->open(nstrm, fd, mode);
+    return othr_netio ? (othr_netio->open)(nstrm, fd, mode) :
+      (core_othr_netio->open)(nstrm, fd, mode);
   }
 
   destroy_pool(nstrm->strm_pool);
+  nstrm->strm_pool = NULL;
 
   errno = EPERM;
   return NULL;
@@ -399,16 +423,16 @@ pr_netio_stream_t *pr_netio_reopen(pr_netio_stream_t *nstrm, int fd, int mode) {
   }
 
   if (nstrm->strm_type == PR_NETIO_STRM_CTRL)
-    return ctrl_netio ? ctrl_netio->reopen(nstrm, fd, mode) :
-      core_ctrl_netio->reopen(nstrm, fd, mode);
+    return ctrl_netio ? (ctrl_netio->reopen)(nstrm, fd, mode) :
+      (core_ctrl_netio->reopen)(nstrm, fd, mode);
 
   if (nstrm->strm_type == PR_NETIO_STRM_DATA)
-    return data_netio ? data_netio->reopen(nstrm, fd, mode) :
-      core_data_netio->reopen(nstrm, fd, mode);
+    return data_netio ? (data_netio->reopen)(nstrm, fd, mode) :
+      (core_data_netio->reopen)(nstrm, fd, mode);
 
   if (nstrm->strm_type == PR_NETIO_STRM_OTHR)
-    return othr_netio ? othr_netio->reopen(nstrm, fd, mode) :
-      core_othr_netio->reopen(nstrm, fd, mode);
+    return othr_netio ? (othr_netio->reopen)(nstrm, fd, mode) :
+      (core_othr_netio->reopen)(nstrm, fd, mode);
 
   errno = EPERM;
   return NULL;
@@ -451,18 +475,18 @@ int pr_netio_poll(pr_netio_stream_t *nstrm) {
 
     switch (nstrm->strm_type) {
       case PR_NETIO_STRM_CTRL:
-        res = ctrl_netio ? ctrl_netio->poll(nstrm) :
-          core_ctrl_netio->poll(nstrm);
+        res = ctrl_netio ? (ctrl_netio->poll)(nstrm) :
+          (core_ctrl_netio->poll)(nstrm);
         break;
 
       case PR_NETIO_STRM_DATA:
-        res = data_netio ? data_netio->poll(nstrm) :
-          core_data_netio->poll(nstrm);
+        res = data_netio ? (data_netio->poll)(nstrm) :
+          (core_data_netio->poll)(nstrm);
         break;
 
       case PR_NETIO_STRM_OTHR:
-        res = othr_netio ? othr_netio->poll(nstrm) :
-          core_othr_netio->poll(nstrm);
+        res = othr_netio ? (othr_netio->poll)(nstrm) :
+          (core_othr_netio->poll)(nstrm);
         break;
     }
 
@@ -508,16 +532,16 @@ int pr_netio_postopen(pr_netio_stream_t *nstrm) {
   }
 
   if (nstrm->strm_type == PR_NETIO_STRM_CTRL)
-    return ctrl_netio ? ctrl_netio->postopen(nstrm) :
-      core_ctrl_netio->postopen(nstrm);
+    return ctrl_netio ? (ctrl_netio->postopen)(nstrm) :
+      (core_ctrl_netio->postopen)(nstrm);
 
   if (nstrm->strm_type == PR_NETIO_STRM_DATA)
-    return data_netio ? data_netio->postopen(nstrm) :
-      core_data_netio->postopen(nstrm);
+    return data_netio ? (data_netio->postopen)(nstrm) :
+      (core_data_netio->postopen)(nstrm);
 
   if (nstrm->strm_type == PR_NETIO_STRM_OTHR)
-    return othr_netio ? othr_netio->postopen(nstrm) :
-      core_othr_netio->postopen(nstrm);
+    return othr_netio ? (othr_netio->postopen)(nstrm) :
+      (core_othr_netio->postopen)(nstrm);
 
   errno = EPERM;
   return -1;
@@ -584,22 +608,26 @@ int pr_netio_write(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
         /* We have to potentially restart here as well, in case we get EINTR. */
         do {
           pr_signals_handle(); 
+
+          if (XFER_ABORTED)
+            break;
+
           run_schedule();
 
           switch (nstrm->strm_type) {
             case PR_NETIO_STRM_CTRL:
-              bwritten = ctrl_netio ? ctrl_netio->write(nstrm, buf, buflen) :
-                core_ctrl_netio->write(nstrm, buf, buflen);
+              bwritten = ctrl_netio ? (ctrl_netio->write)(nstrm, buf, buflen) :
+                (core_ctrl_netio->write)(nstrm, buf, buflen);
                 break;
 
             case PR_NETIO_STRM_DATA:
-              bwritten = data_netio ? data_netio->write(nstrm, buf, buflen) :
-                core_data_netio->write(nstrm, buf, buflen);
+              bwritten = data_netio ? (data_netio->write)(nstrm, buf, buflen) :
+                (core_data_netio->write)(nstrm, buf, buflen);
               break;
 
             case PR_NETIO_STRM_OTHR:
-              bwritten = othr_netio ? othr_netio->write(nstrm, buf, buflen) :
-                core_othr_netio->write(nstrm, buf, buflen);
+              bwritten = othr_netio ? (othr_netio->write)(nstrm, buf, buflen) :
+                (core_othr_netio->write)(nstrm, buf, buflen);
               break;
           }
 
@@ -644,22 +672,30 @@ int pr_netio_write_async(pr_netio_stream_t *nstrm, char *buf, size_t buflen) {
 
   while (buflen) {
     do {
+
+      /* Do NOT check for XFER_ABORTED here.  After a client aborts a
+       * transfer, proftpd still needs to send the 426 response code back
+       * to the client via the control connection; checking for XFER_ABORTED
+       * here would block that response code sending, which in turn causes
+       * problems for clients waiting for that response code.
+       */
+
       pr_signals_handle();
 
       switch (nstrm->strm_type) {
         case PR_NETIO_STRM_CTRL:
-          bwritten = ctrl_netio ? ctrl_netio->write(nstrm, buf, buflen) :
-            core_ctrl_netio->write(nstrm, buf, buflen);
+          bwritten = ctrl_netio ? (ctrl_netio->write)(nstrm, buf, buflen) :
+            (core_ctrl_netio->write)(nstrm, buf, buflen);
           break;
 
         case PR_NETIO_STRM_DATA:
-          bwritten = data_netio ? data_netio->write(nstrm, buf, buflen) :
-            core_data_netio->write(nstrm, buf, buflen);
+          bwritten = data_netio ? (data_netio->write)(nstrm, buf, buflen) :
+            (core_data_netio->write)(nstrm, buf, buflen);
           break;
 
         case PR_NETIO_STRM_OTHR:
-          bwritten = othr_netio ? othr_netio->write(nstrm, buf, buflen) :
-            core_othr_netio->write(nstrm, buf, buflen);
+          bwritten = othr_netio ? (othr_netio->write)(nstrm, buf, buflen) :
+            (core_othr_netio->write)(nstrm, buf, buflen);
           break;
       }
 
@@ -708,6 +744,7 @@ int pr_netio_read(pr_netio_stream_t *nstrm, char *buf, size_t buflen,
 
   while (bufmin > 0) {
     polling:
+
     switch (pr_netio_poll(nstrm)) {
       case 1:
         return -2;
@@ -718,22 +755,26 @@ int pr_netio_read(pr_netio_stream_t *nstrm, char *buf, size_t buflen,
       default:
         do {
           pr_signals_handle();
+
+          if (XFER_ABORTED)
+            break;
+
           run_schedule();
 
           switch (nstrm->strm_type) {
             case PR_NETIO_STRM_CTRL:
-              bread = ctrl_netio ? ctrl_netio->read(nstrm, buf, buflen) :
-                core_ctrl_netio->read(nstrm, buf, buflen);
+              bread = ctrl_netio ? (ctrl_netio->read)(nstrm, buf, buflen) :
+                (core_ctrl_netio->read)(nstrm, buf, buflen);
                 break;
 
             case PR_NETIO_STRM_DATA:
-              bread = data_netio ? data_netio->read(nstrm, buf, buflen) :
-                core_data_netio->read(nstrm, buf, buflen);
+              bread = data_netio ? (data_netio->read)(nstrm, buf, buflen) :
+                (core_data_netio->read)(nstrm, buf, buflen);
               break;
 
             case PR_NETIO_STRM_OTHR:
-              bread = othr_netio ? othr_netio->read(nstrm, buf, buflen) :
-                core_othr_netio->read(nstrm, buf, buflen);
+              bread = othr_netio ? (othr_netio->read)(nstrm, buf, buflen) :
+                (core_othr_netio->read)(nstrm, buf, buflen);
               break;
           }
 
@@ -775,20 +816,20 @@ int pr_netio_shutdown(pr_netio_stream_t *nstrm, int how) {
   }
 
   if (nstrm->strm_type == PR_NETIO_STRM_CTRL) {
-    res = ctrl_netio ? ctrl_netio->shutdown(nstrm, how) :
-      core_ctrl_netio->shutdown(nstrm, how);
+    res = ctrl_netio ? (ctrl_netio->shutdown)(nstrm, how) :
+      (core_ctrl_netio->shutdown)(nstrm, how);
     return res;
   }
 
   if (nstrm->strm_type == PR_NETIO_STRM_DATA) {
-    res = data_netio ? data_netio->shutdown(nstrm, how) :
-      core_data_netio->shutdown(nstrm, how);
+    res = data_netio ? (data_netio->shutdown)(nstrm, how) :
+      (core_data_netio->shutdown)(nstrm, how);
     return res;
   }
 
   if (nstrm->strm_type == PR_NETIO_STRM_OTHR) {
-    res = othr_netio ? othr_netio->shutdown(nstrm, how) :
-      core_othr_netio->shutdown(nstrm, how);
+    res = othr_netio ? (othr_netio->shutdown)(nstrm, how) :
+      (core_othr_netio->shutdown)(nstrm, how);
     return res;
   }
 
@@ -800,6 +841,12 @@ char *pr_netio_gets(char *buf, size_t buflen, pr_netio_stream_t *nstrm) {
   char *bp = buf;
   int toread;
   pr_buffer_t *pbuf = NULL;
+
+  if (buflen == 0) {
+    errno = EINVAL;
+    return NULL;
+  }
+
   buflen--;
 
   if (nstrm->strm_buf)
@@ -866,6 +913,12 @@ char *pr_netio_telnet_gets(char *buf, size_t buflen,
   static unsigned char mode = 0;
   int toread;
   pr_buffer_t *pbuf = NULL;
+
+  if (buflen == 0) {
+    errno = EINVAL;
+    return NULL;
+  }
+
   buflen--;
 
   if (in_nstrm->strm_buf)
