@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2007 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,7 +27,7 @@
 /* Various basic support routines for ProFTPD, used by all modules
  * and not specific to one or another.
  *
- * $Id: support.c,v 1.78 2005/09/28 02:06:26 castaglia Exp $
+ * $Id: support.c,v 1.92 2007/02/15 17:54:58 castaglia Exp $
  */
 
 #include "conf.h"
@@ -321,10 +321,9 @@ char *dir_realpath(pool *p, const char *path) {
   return pstrdup(p, buf);
 }
 
-/* Takes a directory and returns it's absolute version.  ~username
- * references are appropriately interpolated.  "Absolute" includes
- * a *full* reference based on the root directory, not upon a chrooted
- * dir.
+/* Takes a directory and returns its absolute version.  ~username references
+ * are appropriately interpolated.  "Absolute" includes a _full_ reference
+ * based on the root directory, not upon a chrooted dir.
  */
 char *dir_abs_path(pool *p, const char *path, int interpolate) {
   char *res = NULL;
@@ -336,16 +335,27 @@ char *dir_abs_path(pool *p, const char *path, int interpolate) {
     return NULL;
 
   if (*path != '/') {
-    if (session.chroot_path)
+    if (session.chroot_path) {
       res = pdircat(p, session.chroot_path, pr_fs_getcwd(), path, NULL);
-    else
+
+    } else {
       res = pdircat(p, pr_fs_getcwd(), path, NULL);
+    }
 
-  } else if (session.chroot_path)
-    res = pdircat(p, session.chroot_path, path, NULL);
-
-  else
-    res = pstrdup(p, path);
+  } else {
+    if (session.chroot_path) {
+      if (strncmp(path, session.chroot_path,
+          strlen(session.chroot_path)) != 0) {
+        res = pdircat(p, session.chroot_path, path, NULL);
+ 
+      } else {
+        res = pstrdup(p, path);
+      }
+ 
+    } else {
+      res = pstrdup(p, path);
+    }
+  }
 
   return res;
 }
@@ -725,7 +735,11 @@ char *sreplace(pool *p, char *s, ...) {
 	if (((cp + rlen) - pbuf + 1) > blen) {
 	  pr_log_pri(PR_LOG_ERR,
 		  "WARNING: attempt to overflow internal ProFTPD buffers");
-	  cp = pbuf + blen - 1;
+	  cp = pbuf;
+          if (blen >= BUFSIZ)
+            blen = BUFSIZ;
+          cp += (blen - 1);
+
 	  goto done;
 
 	} else {
@@ -741,7 +755,11 @@ char *sreplace(pool *p, char *s, ...) {
       if ((cp - pbuf + 1) >= blen) {
 	pr_log_pri(PR_LOG_ERR,
 		"WARNING: attempt to overflow internal ProFTPD buffers");
-	cp = pbuf + blen - 1;
+	cp = pbuf;
+        if (blen >= BUFSIZ)
+          blen = BUFSIZ;
+        cp += (blen - 1);
+
 	goto done;
       }
       *cp++ = *src++;
@@ -767,8 +785,15 @@ char *sreplace(pool *p, char *s, ...) {
 unsigned char memscrub_ctr = 0;
 
 void pr_memscrub(void *ptr, size_t ptrlen) {
-  unsigned char *p = ptr;
-  size_t loop = ptrlen;
+  unsigned char *p;
+  size_t loop;
+
+  if (!ptr || ptrlen == 0) {
+    return;
+  }
+
+  p = ptr;
+  loop = ptrlen;
 
   while (loop--) {
     *(p++) = memscrub_ctr++;
@@ -785,6 +810,11 @@ void pr_memscrub(void *ptr, size_t ptrlen) {
 char *sstrcat(char *dest, const char *src, size_t n) {
   register char *d;
 
+  if (!dest || !src || n == 0) {
+    errno = EINVAL;
+    return NULL;
+  }
+
   for (d = dest; *d && n > 1; d++, n--) ;
 
   while (n-- > 1 && *src)
@@ -807,9 +837,58 @@ struct tm *pr_gmtime(pool *p, const time_t *t) {
 struct tm *pr_localtime(pool *p, const time_t *t) {
   struct tm *sys_tm, *dup_tm;
 
+#ifdef HAVE_TZNAME
+  char *tzname_dup[2];
+
+  /* The localtime(3) function has a nasty habit of changing the tzname
+   * global variable as a side-effect.  This can cause problems, as when
+   * the process has become chrooted, and localtime(3) sets/changes
+   * tzname wrong.  (For more information on the tzname global variable,
+   * see the tzset(3) man page.)
+   *
+   * The best way to deal with this issue (which is especially prominent
+   * on systems running glibc-2.3 or later, which is particularly ill-behaved
+   * in a chrooted environment, as it assumes the ability to find system
+   * timezone files at paths which are no longer valid within the chroot)
+   * is to set the TZ environment variable explicitly, before starting
+   * proftpd.  You can also use the SetEnv configuration directive within
+   * the proftpd.conf to set the TZ environment variable, e.g.:
+   *
+   *  SetEnv TZ PST
+   *
+   * To try to help sites which fail to do this, the tzname global variable
+   * will be copied prior to the localtime(3) call, and the copy restored
+   * after the call.  (Note that calling the ctime(3) and mktime(3)
+   * functions also causes a similar overwriting/setting of the tzname
+   * environment variable.)
+   *
+   * This hack is also used in the lib/pr-syslog.c code, to work around
+   * mktime(3) antics.
+   */
+  memcpy(&tzname_dup, tzname, sizeof(tzname_dup));
+#endif /* HAVE_TZNAME */
+
   sys_tm = localtime(t);
-  dup_tm = pcalloc(p, sizeof(struct tm));
-  memcpy(dup_tm, sys_tm, sizeof(struct tm));
+
+  if (p) {
+    /* If the caller provided a pool, make a copy of the returned
+     * struct tm, allocated out of that pool.
+     */
+    dup_tm = pcalloc(p, sizeof(struct tm));
+    memcpy(dup_tm, sys_tm, sizeof(struct tm));
+
+  } else {
+
+    /* Other callers do not require pool-allocated copies, and instead
+     * are happy with the struct tm as is.
+     */
+    dup_tm = sys_tm;
+  }
+
+#ifdef HAVE_TZNAME
+  /* Restore the old tzname values prior to returning. */
+  memcpy(tzname, tzname_dup, sizeof(tzname_dup));
+#endif /* HAVE_TZNAME */
 
   return dup_tm;
 }
@@ -823,7 +902,7 @@ const char *pr_strtime(time_t t) {
 
   memset(buf, '\0', sizeof(buf));
 
-  tr = localtime(&t);
+  tr = pr_localtime(NULL, &t);
   if (tr != NULL) {
     snprintf(buf, sizeof(buf), "%s %s %2d %02d:%02d:%02d %d",
       days[tr->tm_wday], mons[tr->tm_mon], tr->tm_mday, tr->tm_hour,
