@@ -23,7 +23,7 @@
  */
 
 /* Table API implementation
- * $Id: table.c,v 1.7 2005/06/16 21:27:27 castaglia Exp $
+ * $Id: table.c,v 1.9 2006/05/26 16:43:15 castaglia Exp $
  */
 
 #include "conf.h"
@@ -961,6 +961,16 @@ int pr_table_ctl(pr_table_t *tab, int cmd, void *arg) {
   return -1;
 }
 
+void *pr_table_pcalloc(pr_table_t *tab, size_t sz) {
+  if (!tab ||
+      sz == 0) {
+    errno = EINVAL;
+    return NULL;
+  }
+
+  return pcalloc(tab->pool, sz);
+}
+
 void pr_table_dump(void (*dumpf)(const char *fmt, ...), pr_table_t *tab) {
   register unsigned int i;
 
@@ -998,8 +1008,8 @@ void pr_table_dump(void (*dumpf)(const char *fmt, ...), pr_table_t *tab) {
     while (ent) {
       pr_signals_handle();
 
-      dumpf("[chain %u#%u] '%s' => '%s'", i, j++, ent->key->key_data,
-        ent->value_data);
+      dumpf("[chain %u#%u] '%s' => '%s' (%u)", i, j++, ent->key->key_data,
+        ent->value_data, ent->value_datasz);
       ent = ent->next;
     }
   }
