@@ -1,7 +1,7 @@
 /*
  * ProFTPD - ftptop: a utility for monitoring proftpd sessions
  * Copyright (c) 2000-2002 TJ Saunders <tj@castaglia.org>
- * Copyright (c) 2003 The ProFTPD Project team
+ * Copyright (c) 2003-2006 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,7 +26,7 @@
 /* Shows who is online via proftpd, in a manner similar to top.  Uses the
  * scoreboard files.
  *
- * $Id: ftptop.c,v 1.29 2004/03/05 03:24:36 castaglia Exp $
+ * $Id: ftptop.c,v 1.33 2007/05/22 21:21:25 castaglia Exp $
  */
 
 #define FTPTOP_VERSION "ftptop/0.9"
@@ -42,10 +42,12 @@ static const char *program = "ftptop";
 
 /* ncurses is preferred...*/
 
-#if defined(HAVE_NCURSES_H) && defined(HAVE_LIBNCURSES) && defined(USE_NCURSES)
+#if defined(HAVE_NCURSES_H) && defined(HAVE_LIBNCURSES) && \
+    defined(PR_USE_NCURSES)
 # define HAVE_NCURSES 1
 # include <ncurses.h>
-#elif defined(HAVE_CURSES_H) && defined(HAVE_LIBCURSES) && defined(USE_CURSES)
+#elif defined(HAVE_CURSES_H) && defined(HAVE_LIBCURSES) && \
+    defined(PR_USE_CURSES)
 # define HAVE_CURSES 1
 /* Sigh...portability.  It seems that Solaris' curses.h (at least for 2.8)
  * steps on wide-character macros, generating compiler warnings.  This, then
@@ -66,8 +68,8 @@ static const char *program = "ftptop";
 #define FTPTOP_REG_DISPLAY_FMT	"%-5u %s %-8.8s %-20.20s %-15s %-6.6s %4s %-*.*s\n"
 
 /* These are for displaying tranfer data: "PID S USER CLIENT KB/s %DONE" */
-#define FTPTOP_XFER_HEADER_FMT	"%-5s %s %-8s %-20s %-5s    %-*s\n"
-#define FTPTOP_XFER_DISPLAY_FMT	"%-5u %s %-8.8s %-46.46s %-3.2f %6s%%s\n"
+#define FTPTOP_XFER_HEADER_FMT	"%-5s %s %-8s %-44s %-10s %-*s\n"
+#define FTPTOP_XFER_DISPLAY_FMT	"%-5u %s %-8.8s %-44.44s %-10.2f %-*s\n"
 
 #define FTPTOP_REG_ARG_MIN_SIZE		20
 #define FTPTOP_XFER_DONE_MIN_SIZE	6
@@ -88,7 +90,7 @@ static const char *program = "ftptop";
 static int delay = 2;
 static unsigned int display_mode = FTPTOP_SHOW_REG;
 
-static char *config_filename = CONFIG_FILE_PATH;
+static char *config_filename = PR_CONFIG_FILE_PATH;
 
 /* Scoreboard variables */
 static time_t ftp_uptime = 0;
@@ -200,8 +202,10 @@ static const char *show_ftpd_uptime(void) {
 
   updays = (int) uptime_secs / (60 * 60 * 24);
 
-  if (updays)
-    pos += sprintf(buf + pos, "%d day%s, ", updays, (updays != 1) ? "s" : "");
+  if (updays) {
+    pos += snprintf(buf + pos, sizeof(buf) - pos, "%d day%s, ", updays,
+      (updays != 1) ? "s" : "");
+  }
 
   upminutes = (int) uptime_secs / 60;
 
@@ -210,11 +214,13 @@ static const char *show_ftpd_uptime(void) {
 
   upminutes = upminutes % 60;
 
-  if (uphours)
-    pos += sprintf(buf + pos, "%2d hr%s %02d min", uphours,
+  if (uphours) {
+    pos += snprintf(buf + pos, sizeof(buf) - pos, "%2d hr%s %02d min", uphours,
       (uphours != 1) ? "s" : "", upminutes);
-  else
-    pos += sprintf(buf + pos, "%d min", upminutes);
+
+  } else {
+    pos += snprintf(buf + pos, sizeof(buf) - pos, "%d min", upminutes);
+  }
 
   return buf;
 }
@@ -314,10 +320,20 @@ static void process_opts(int argc, char *argv[]) {
           exit(1);
         }
 
+        if (delay > 15) {
+          fprintf(stderr, "%s: delay of 0-15 seconds only supported\n",
+            program);
+          exit(1);
+        }
+
         break;
 
       case 'f':
-        util_set_scoreboard(optarg);
+        if (util_set_scoreboard(optarg) < 0) {
+          fprintf(stderr, "%s: unable to use scoreboard '%s': %s\n",
+            program, optarg, strerror(errno));
+          exit(1);
+        }
         break;
 
       case 'h':
@@ -441,6 +457,7 @@ static void read_scoreboard(void) {
         (unsigned int) score->sce_pid, status, score->sce_user,
         score->sce_client_name,
         (score->sce_xfer_len / 1024.0) / (score->sce_xfer_elapsed / 1000),
+        FTPTOP_XFER_DONE_SIZE,
         *status == 'D' ?
           calc_percent_done(score->sce_xfer_size, score->sce_xfer_done) :
           "(n/a)");
